class CERROR_MESSAGES:
    DESCRIPTION_WRONG_TYPE: str = "description must be of type string"
    ALL_ARGUMENTS_MICROCONTROLLER: str = "all arguments must be of type Microcontroller"
    ALL_ARGUMENTS_DEVICE: str = "all arguments must be of type Device"
    ITEM_ALREADY_EXISTS_IN_DICT: str = "Key already exists in dictionary: "
    CLASS_NO_INIT: str = "This class should not be instantiated."
    PACKET_NOT_SET: str = "method, node_identity or payload are None. set these properties before calling serialise."
    NODE_IDENTITY_LENGTH: str = "a node identity must be 4 characters long"
    TYPE_MUST_BE_LIST: str = "the argument must be of type list"
    TYPE_MUST_BE_CALLABLE: str = "callback must be of type function"

class CCOMMAND:
    RECORD_CREATE: str = "RECORD_CREATE"
    RECORD_START: str = "RECORD_START"
    RECORD_STOP: str = "RECORD_STOP"
    RECORD_READ: str = "RECORD_READ"
    
    RESTART: str = "RESTART"
    BATTERY_READ: str = "BATTERY_READ"
    IDENTIFY: str = "IDENTIFY"

    CONNECTION_READ: str = "CONNECTION_READ"
    WIFI_PROFILE_CREATE: str = "WIFI_PROFILE_CREATE"
    WIFI_PROFILE_READ: str = "WIFI_PROFILE_READ"
    WIFI_PROFILE_ACTIVE_READ: str = "WIFI_PROFILE_ACTIVE_READ"
    WIFI_PROFILE_ALL_READ: str = "WIFI_PROFILE_ALL_READ"
    WIFI_PROFILE_ACTIVE_SELECT: str = "WIFI_PROFILE_SELECT"
    WIFI_PROFILE_DELETE: str = "WIFI_PROFILE_DELETE"

#custom commands are transfered with RECORD_READ

class CFEEDBACK:
    RECORD_INFO: str = "RECORD_INFO"
    BATTERY_INFO: str = "BATTERY_INFO"
    IDENTIFY_INFO: str = "IDENTIFY_INFO"
    WIFI_INFO: str = "WIFI_INFO"
    WIFI_PROFILE_INFO: str = "WIFI_PROFILE_INFO"
    PROTOCOL_INFO: str = "PROTOCOL_INFO"
    CAPABILITIES_INFO: str = "CAPABILITIES_INFO"

class CNETWORK:
    SERIAL_NAME: str = "SERIAL"
    WIFI_NAME: str = "WIFI"
    BLE_NAME: str = "BLE"
    OUT_OF_ORDER_PACKET_MAX_SIZE: int = 5
    SEQUENCE_MAX_NUMBER_SIZE: int = 65535 #2¹⁶-1 unsigned
    SEND_ACK_PACKETS: bool = True
    HEARTBEAT_INTERVALint = 1000 #ms
    BAUDRATE: int = 115200
    TIMEOUT = 1
    BUFFER_SIZE = 1000
    SERIAL_ARDUINO_WORDS = ["Arduino", "CP2104"] #add more words if necessary to find other boards
    SELECTED_PATHS = ["platformio.ini", "lib", "src"]
    BUSY_WAITING_TIMEOUT = 0.001
    PORT_ALREADY_USED_CODE = 98

class CPACKET:
    HEADER_SIZE: int = 10
    HEADER_PAYLOADSIZE_POSITION: int = 8 #plus 9 because payload takes 2 bytes
    #reserve the 6 characters in ASCII table starting with 0x20 and ending with 0x26 as the beginning byte of our packets
    METHOD_ACKNOWLEDGEMENT: int = 0x20 #SP #32
    METHOD_DATA: int = 0x21 #! #33
    METHOD_COMMAND: int = 0x22 #" #34
    METHOD_HEARTBEAT: int = 0x23 ## #35
    METHOD_DEBUG: int = 0x24 #$ #36
    METHOD_INFO: int = 0x25 #% #37
    METHOD_ERROR: int = 0x26 #& 38
    HEADER_ENCODING: str = "ascii"
    PAYLOAD_ENCODING: str = "utf-8"
    HEADER_FORMAT_CHARACTERS: str = "!BIHBH"
    NODE_IDENTITY_LENGTH: int = 4

    CRC_WIDTH: int = 8
    CRC_POLYNOMIAL: int = 0xa7
    CRC_INIT_VALUE: int = 0x00
    CRC_FINAL_XOR_VALUE: int = 0x00
    CRC_REVERSE_INPUT: bool = True
    CRC_REVERSE_OUTPUT: bool = True

class CCAPABILITIES:
    DEFAULT_DURATION: int = 0#milli seconds
    DEFAULT_MAX_SAMPLES: int = 0
    DEFAULT_DELAY: int = 0#milli seconds
    DEFAULT_INCLUDE_SEQUENCE: bool = False
    DEFAULT_INCLUDE_TIMESTAMP: bool = False
    DEFAULT_INCLUDE_SENSOR: bool = False
    DEFAULT_INCLUDE_ACTUATOR: bool = False
    DEFAULT_SAMPLE_RATE: int = 250
    DEFAULT_DATA_ON_CHANGE: bool = False

class CCLIENT:
    PORT = 7777
    CONFIG_FILE_NAME = "mayako.json"
    DEFAULT_ARDUINO_FOLDER = "arduino"

class CGENERAL:
    DEBUG = 1


import uuid

def get_macaddress() -> str:
    """
    returns the macaddress of the underlying computer.
    
    this function returns the macaddress of the underlying computer. it uses the uuid standard library package. the format is e.g. 02:73:ab:ab:0e:e7. it returns all 1 if no mac address was found.

    References:
        https://stackoverflow.com/a/28928058

    Returns:
        mac (str):
            Example: 02:73:ab:ab:0e:e7
    """
    address_as_int = uuid.getnode()
    h = iter(hex(address_as_int)[2:].zfill(12))
    mac = ":".join(i + next(h) for i in h)

    return mac


from enum import Enum
from typing import List, Dict, Union, Optional, TypeVar, Generic

from ..Network.IntegrityMiddleware import IntegrityMiddleware
from ..Network.NetworkInterface import NetworkInterface
from ..Network.BLE import BLE
from ..Network.WiFiUDP import WiFiUDP
from ..Network.Serial import Serial
from ..Utils.Logger import LoggerInstance, LoggerType
from ..Models.SerialInfo import SerialInfo
from ..Models.BLEInfo import BLEInfo
from ..Models.WiFiProfile import WiFiProfile
from ..Network.Packet import Packet
from ..Network.NetworkInterface import NetworkProtocols

#https://stackoverflow.com/a/51976841

class RecordStatus(Enum):
    RECORD_INIT=0
    PROTOCOL_DETERMINED=1
    CAPABILITY_UPDATED=2
    RECORD_STARTED=3
    STOP_SIGNALED=4
    RECORD_STOPPED=5

class NetworkManager:

    _logger: LoggerType
    _mc_identity: str
    _network_protocol: NetworkProtocols
    _record_status: RecordStatus
    _network_implementation: NetworkInterface
    #_integrity_middelware: IntegrityMiddleware

    def __init__(self, mc_identity: str) -> None:
        self._logger = LoggerInstance.get()
        self._mc_identity = mc_identity
        self._network_protocol = None
        self._record_status = RecordStatus.RECORD_INIT
        self._network_implementation = None
        #self._integrity_middelware = IntegrityMiddleware(self._protocol)

    def use_serial(self, serial_address: SerialInfo) -> None:
        self._network_protocol = NetworkProtocols.SERIAL
        self._serial_address = serial_address
        self._network_implementation = Serial(serial_port=self._serial_address.port)

    def use_ble(self, ble_address: BLEInfo) -> None:
        self._network_protocol = NetworkProtocols.BLE
        self._ble_address = ble_address
        self._network_implementation = BLE(address=self._ble_address.address, service_uuid=self._ble_address.service_uuid, characteristic_uuid=self._ble_address.characteristic_uuid)

    def get_mc_identity(self) -> str:
        return self._mc_identity

    def use_wifi(self, wifi_profile: WiFiProfile) -> None:
        raise NotImplementedError
    
    def connect(self) -> None:
        self._network_implementation.connect()
    
    def send_data(self, packet: Packet) -> None:
        self._network_implementation.write(packet)
    
    def read_data(self) -> Optional[Packet]:
        return self._network_implementation.read()
    
    def disconnect(self) -> None:
        self._network_implementation.disconnect()

    def check_connection(self) -> bool:
        return self._network_implementation.check_connection()
    
    def get_protocol_name(self) -> str:
        return self._network_implementation.get_protocol_name()


#this is the contrast to LEDActuator. it handles custom execution mode while providing nearly not syntax highlighting
from mayako.MayakoData import MayakoData
from .Actuator import Actuator


class CustomActuator(Actuator):

    def __init__(self, identity: str, model: MayakoData) -> None:
        super().__init__(identity, model)

    def execute(self, cmd_name: str, parameters: dict = {}) -> None:
        #add mc_id
        #add id
        #self.commands.execute(mc_id, act_id, cmd_name)
        pass


import tkinter as tk
from tkinter import messagebox

from ..Views.MainView import MainView
from ..Views.AddWiFiProfileView import AddWiFiProfileView
from ...Models.WiFiProfile import WiFiProfile
from ...MayakoData import MayakoData
from ...Client import Client

class AddWiFiProfileController:

    main_view: MainView
    add_wifi_frame: AddWiFiProfileView

    def __init__(self, main_view: MainView, model: MayakoData, client: Client) -> None:
        self.main_view = main_view
        self.model = model
        self.client = client
        self.add_wifi_frame = self.main_view.frame_classes["add_wifi"]
        self._bind()

    def _bind(self) -> None:
        self.add_wifi_frame.return_btn.config(command=self._return_to_main)
        self.add_wifi_frame.add_wifi_profile_btn.config(command=self._add_wifi_profile)

    def _return_to_main(self) -> None:
        self.main_view.switch("home")
        self.add_wifi_frame.wifi_key_entry.delete(0, tk.END)
        self.add_wifi_frame.ssid_entry.delete(0, tk.END)
        self.add_wifi_frame.password_entry.delete(0, tk.END)
        self.add_wifi_frame.client_ip_entry.delete(0, tk.END)
        self.add_wifi_frame.client_port_entry.delete(0, tk.END)

    def _add_wifi_profile(self) -> None:
        wifi_key = self.add_wifi_frame.wifi_key_entry.get()
        ssid = self.add_wifi_frame.ssid_entry.get()
        password = self.add_wifi_frame.password_entry.get()
        client_ip = self.add_wifi_frame.client_ip_entry.get()
        client_port = self.add_wifi_frame.client_port_entry.get()

        if len(wifi_key) < 4:
            messagebox.showerror(title="Error", message="wifi key must be 4 characters long")
            return
        if self.model.get_wifi_profile_by_wifi_key(wifi_key=wifi_key):
            messagebox.showerror(title="Error", message="a wifi profile with this identity already exists. please use another wifi key.")
            return
        if ssid == "":
            messagebox.showerror(title="Error", message="ssid may not be empty")
            return
        if password == "":
            messagebox.showerror(title="Error", message="password may not be empty")
            return
        if client_ip == "":
            messagebox.showerror(title="Error", message="client ip may not be empty")
            return
        try:
            client_port = int(client_port)                
        except ValueError:
            messagebox.showerror(title="Error", message="client port must be an integer and may not be empty")
            return

        profile = WiFiProfile(
            wifi_key=wifi_key,
            ssid=ssid,
            password=password,
            client_ip=client_ip,
            client_port=client_port
            )
        
        self.model.add_wifi_profile([profile])

        self.main_view.switch("home")


from ..Views.MainView import MainView
from .AddDeviceController import AddDeviceController
from .DeviceDetailsController import DeviceDetailsController
from .AddWiFiProfileController import AddWiFiProfileController
from .EditWiFiProfileController import EditWiFiProfileController
from .DownloadArduinoController import DownloadArduinoController
from .HomeController import HomeController
from ...MayakoData import MayakoData
from ...Client import Client

class MainController:

    def __init__(self, main_view: MainView, model: MayakoData, client: Client) -> None:
        self.model = model
        self.main_view = main_view
        self.client = client
        self.home_controller = HomeController(self.main_view, self.model, self.client)
        self.add_device_controller = AddDeviceController(self.main_view, self.model, self.client)
        self.add_wifi_profile_controller = AddWiFiProfileController(self.main_view, self.model, self.client)
        self.edit_wifi_profile_controller = EditWiFiProfileController(self.main_view, self.model, self.client)
        self.device_deails_controller = DeviceDetailsController(self.main_view, self.model, self.client)
        self.download_arduino_controller = DownloadArduinoController(self.main_view, self.model, self.client)

    def start(self) -> None:
        self.main_view.switch("home")
        self.main_view.start_mainloop()




from tkinter import messagebox
import tkinter as tk
import threading

from ..Views.MainView import MainView
from ..Views.AddDeviceView import AddDeviceView
from ...Network.NetworkInterface import NetworkProtocols
from ...Service.SerialScanner import scan_serial_ports
from ...Service.ArduinoUploader import ArduinoUploader
from ...MayakoData import MayakoData
from ...Devices.MicroController import MicroControllerCapabilities
from ...Config import CNETWORK
from ...Client import Client
from ...Models.WiFiProfile import WiFiProfile
from ...Models.BLEInfo import BLEInfo

class AddDeviceController:
    """
    this class adds a new mayako-node device to the mayako-core project.
    """
    main_view: MainView
    add_device_frame: AddDeviceView
    uploading: bool

    def __init__(self, main_view: MainView, model: MayakoData, client: Client) -> None:
        self.main_view = main_view
        self.model = model
        self.client = client
        self.add_device_frame = self.main_view.frame_classes["add_device"]#type: ignore
        self._bind()
        self._model_subscribe()
        self.uploading = False

        self._load_wifi_profiles_into_list()
        self._load_serial_ports_into_list()

    def _load_wifi_profiles_into_list(self) -> None:
        self.add_device_frame.wifi_profile_listbox.delete(0, tk.END)

        for key, wifi_profile in enumerate(self.model.get_wifi_profiles()):
            self.add_device_frame.wifi_profile_listbox.insert(key, wifi_profile.format_string())

    def _load_serial_ports_into_list(self) -> None:
        self.add_device_frame.serial_port_listbox.delete(0, tk.END)

        for key, serial_port in enumerate(self.model.get_serial_ports()):
            self.add_device_frame.serial_port_listbox.insert(key, serial_port.format_string())
            
    def _bind(self) -> None:
        self.add_device_frame.return_btn.config(command=self._return_to_main)
        self.add_device_frame.ble_check.config(command=self._selected_ble)
        self.add_device_frame.wifi_check.config(command=self._selected_wifi)
        self.add_device_frame.flash_device_btn.config(command=self._flash_device)
        self.add_device_frame.wifi_var.trace_add("write", callback=self._toggle_wifi_profiles)
        self.add_device_frame.serial_devices_update_btn.config(command=self._udpate_serial_devices)

    def _model_subscribe(self) -> None:
        self.model.subscribe("wifi_profiles_update", self._update_wifi_profiles)
        self.model.subscribe("serial_ports_update", self._update_serial_ports)

    def _update_wifi_profiles(self, *_) -> None:
        self._load_wifi_profiles_into_list()

    def _update_serial_ports(self, *_) -> None:
        self._load_serial_ports_into_list()

    def _return_to_main(self) -> None:
        self.main_view.switch("home")
        self.add_device_frame.ble_var.set(False)
        self.add_device_frame.wifi_var.set(False)
        self.add_device_frame.identity_entry.delete(0, tk.END)
        self.add_device_frame.status_label.config(text=self.add_device_frame.status_text[0][0])
        self.add_device_frame.status_label.config(fg=self.add_device_frame.status_text[0][1])
        self.uploading = False
        self.serial_ports = []
        self.add_device_frame.serial_port_listbox.delete(0, tk.END)

    def _selected_ble(self) -> None:
        selected_ble = self.add_device_frame.ble_var.get()
        if selected_ble:
            self.add_device_frame.wifi_var.set(False)
    
    def _selected_wifi(self) -> None:
        selected_wifi = self.add_device_frame.wifi_var.get()
        if selected_wifi:
            self.add_device_frame.ble_var.set(False)

    def _toggle_wifi_profiles(self, *_) -> None:
        selected_menu = self.add_device_frame.wifi_var.get()
        if selected_menu:
            self.add_device_frame.wifi_profile_listbox.grid(row=3, column=0, columnspan=2, padx=10, pady=5, sticky="nsew")
            self.add_device_frame.wifi_profile_scrollbar.grid(row=3, column=1, padx=(0, 10), pady=5, sticky="nes")
        else:
            self.add_device_frame.wifi_profile_listbox.grid_remove()
            self.add_device_frame.wifi_profile_scrollbar.grid_remove()

    def _udpate_serial_devices(self) -> None:
        #first clear the list to avoid doubles
        self.add_device_frame.serial_port_listbox.delete(0, tk.END)
        
        #scan for serial ports with arduino description
        self.model.add_serial_port(scan_serial_ports(CNETWORK.SERIAL_ARDUINO_WORDS))
        
        self._load_serial_ports_into_list()

    def _flash_device(self) -> None:
        #prevent uploading multiple times
        if self.uploading: return

        #wireless mode
        wireless_mode = None
        if self.add_device_frame.ble_var.get():
            wireless_mode = NetworkProtocols.BLE

        elif self.add_device_frame.wifi_var.get():
            wireless_mode = NetworkProtocols.WIFI

        else:
            messagebox.showerror(title="Error", message="choose BLE or WiFi")
            return
        
        #wifi profile
        wifi_profile: WiFiProfile = None
        if self.add_device_frame.wifi_var.get():
            wifi_profile_key = self.add_device_frame.wifi_profile_listbox.curselection()

            if not wifi_profile_key:
                messagebox.showerror(title="Error", message="choose a wifi profile")
                return
            
            wifi_profile = self.model.get_wifi_profiles()[wifi_profile_key[0]]

            if not wifi_profile:
                messagebox.showerror(title="Error", message="wifi profile not found")
                return
        
        identity = self.add_device_frame.identity_entry.get()
        if len(identity) != 4:
            messagebox.showerror(title="Error", message="identity must be 4 characters long")
            return
        
        if not identity.isascii():
            messagebox.showerror(title="Error", message="identity must be of type ascii")
            return
        
        if self.model.get_microcontroller_capa_by_identity(identity=identity):
            messagebox.showerror(title="Error", message="a microcontroller with this identity is already in use. please use another identity.")
            return
        
        selected_port_key = self.add_device_frame.serial_port_listbox.curselection()
        if not selected_port_key:
            messagebox.showerror(title="Error", message="select a serial port for upload")
            return

        selected_serial_port = self.model.get_serial_ports()[selected_port_key[0]].port

        #disable flash device button
        self.uploading = True

        #show status in status label - progress
        self.show_progress_status()
        
        #upload device
        uploader = None

        user = self.model.get_this_user()

        #TODO: add to thread?
        if wireless_mode == NetworkProtocols.BLE:
            uploader = ArduinoUploader(user.arduino_folder, identity=identity, serial_port=selected_serial_port, wireless_mode=NetworkProtocols.BLE)
        elif wireless_mode == NetworkProtocols.WIFI:
            uploader = ArduinoUploader(user.arduino_folder, identity=identity, serial_port=selected_serial_port, wireless_mode=NetworkProtocols.WIFI, wifi_profile=wifi_profile)
        else:
            messagebox.showerror(title="Error", message="now upload solution for your settings")
            self.uploading = False
            return
        
        if uploader.check_prerequisits():
            success = uploader.upload()
            if success:
                self.show_success_status()

                if wifi_profile is not None:
                    wifi_key = wifi_profile.wifi_key
                else:
                    wifi_key = None

                new_microcontroller = MicroControllerCapabilities(identity=identity, serial_port=selected_serial_port, protocol=wireless_mode, wifi_key=wifi_key)
                
                self.model.add_microcontroller_capa([new_microcontroller])

                t = threading.Thread(target=self.client._scan_ble_address(mc_identity=new_microcontroller.identity))
                t.start()

            else:
                self.show_error_status()

                messagebox.showerror(title="Error", message="there was an error while uploading")

        else:
            self.show_error_status()
            
            messagebox.showerror(title="Error", message="there was a mistake with the prerequisits for the uploader")
      
        self.uploading = False

    def show_progress_status(self):
        self.add_device_frame.status_label.config(fg=self.add_device_frame.status_text[3][1])
        self.add_device_frame.status_label.config(text=self.add_device_frame.status_text[3][0])
        self.add_device_frame.status_label.update()

    def show_error_status(self):
        self.add_device_frame.status_label.config(text=self.add_device_frame.status_text[1][0])
        self.add_device_frame.status_label.config(fg=self.add_device_frame.status_text[1][1])

    def show_success_status(self):
        self.add_device_frame.status_label.config(text=self.add_device_frame.status_text[2][0])
        self.add_device_frame.status_label.config(fg=self.add_device_frame.status_text[2][1])

    


from tkinter import Frame, Label, Button, Text, Scrollbar, VERTICAL
import tkinter as tk

class DeviceDetailsView(Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.grid_columnconfigure(0, weight=0)
        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(4, weight=1)
        self.grid_rowconfigure(5, weight=1) 

        #headline
        self.headline = Label(self, text="Device Details")
        self.headline.grid(row=0, column=0, columnspan=2, padx=10, pady=10, sticky="w")

        #return button
        self.return_btn = Button(self, text="Return")
        self.return_btn.grid(row=0, column=1, padx=10, pady=10, sticky="e")

        # Identity Label
        self.identity_label = Label(self, text="Identity")
        self.identity_label.grid(row=1, column=0, padx=10, pady=10, sticky="w")
        self.identity_text = Label(self)
        self.identity_text.grid(row=1, column=1, padx=10, pady=10, sticky="w")

        #protocol used label
        self.protocol_label = Label(self, text="Protocol")
        self.protocol_label.grid(row=2, column=0, padx=10, pady=10, sticky="w")
        self.protocol_text = Label(self)
        self.protocol_text.grid(row=2, column=1, padx=10, pady=10, sticky="w")

        #battery
        self.battery_label = Label(self, text="Battery")
        self.battery_label.grid(row=3, column=0, padx=10, pady=10, sticky="w")
        self.battery_text = Label(self)
        self.battery_text.grid(row=3, column=1, padx=10, pady=10, sticky="w")

        #capabiltiies
        self.update_capabilities_btn = Button(self, text="Update Capabilities")
        self.update_capabilities_btn.grid(row=4, column=0, padx=10, pady=5)
        #self.update_capabilities_btn.config(state="disabled")#disabled while device not connected

        #Identify
        #self.identify_device_btn = Button(self, text="Identify Device")
        #self.identify_device_btn.grid(row=4, column=1, padx=10, pady=5, sticky="w")

        #text field for capabiltiies
        self.capabilities_text = Text(self)
        self.capabilities_text.grid(row=5, column=0, columnspan=2, padx=(10, 0), pady=5, sticky="nesw")
    
        self.capabiltiies_scrollbar = Scrollbar(self, orient=VERTICAL, command=self.capabilities_text.yview)
        self.capabiltiies_scrollbar.grid(row=5, column=2, padx=(0, 10), pady=5, sticky="nes")
        self.capabilities_text.config(yscrollcommand=self.capabiltiies_scrollbar.set)

        #label for success, fail, inprogress, idle
        self.status_text = [("", "black"), ("Fail", "red"), ("Success", "green"), ("In Progress", "black")]
        i = 0
        self.status_label = Label(self, text=self.status_text[i][0], fg=self.status_text[i][1])
        self.status_label.grid(row=6, column=1, padx=10, pady=10, sticky="ew")

if __name__ == "__main__":
    root = tk.Tk()
    root.title("Add New WiFi Profile")
    root.geometry("450x600")
    add_wifi_profile_view = DeviceDetailsView(root)
    add_wifi_profile_view.pack(fill="both", expand=True)
    root.mainloop()



from queue import Queue
import json
from typing import Dict

from ..Network.Packet import Packet
from ..Config import CPACKET

class CustomCommands:
    """
    this custom class contains an interface for users to execute custom commands for actuators on the mayako-nodes. directly using this might expose risks, we recommend implementing an actuator class. it uses the clients queue to store packet. these are frequently flushed to the network by the client. the client injects an instance of this class into actuator, which lets it use by the client in a subclass of actuator.
    """

    def __init__(self, client_queue: Queue[Packet]) -> None:
        self._client_queue = client_queue

    def execute(self, mc_identity: str, cmd_name: str, parameters: Dict = {}) -> None:
        doc = parameters
        doc["cmd_name"] = cmd_name

        jsonDoc = json.dumps(doc)

        packet = Packet()
        packet.set_method(CPACKET.METHOD_COMMAND)
        packet.set_node_identity(mc_identity)
        packet.set_payload(jsonDoc)
        
        self._client_queue.put(packet)


from typing import Dict
import tkinter as tk
from tkinter import Frame
from .DownloadArduinoView import DownloadArduinoView
from .AddDeviceView import AddDeviceView
from .AddWiFiProfileView import AddWiFiProfileView
from .EditWiFiProfileView import EditWiFiProfileView
from .DeviceDetailsView import DeviceDetailsView
from .HomeView import HomeView
from .Root import Root

class MainView():

    frame_classes: Dict[str, Root]

    def __init__(self) -> None:
        self.root = Root()
        self.root.bind("<Control-w>", self.stop_mainloop)
        self.frame_classes = {
            "home": HomeView(self.root),
            "edit_wifi": EditWiFiProfileView(self.root),
            "add_wifi": AddWiFiProfileView(self.root),
            "add_device": AddDeviceView(self.root),
            "device_details": DeviceDetailsView(self.root),
            "download": DownloadArduinoView(self.root)
        }
        self.current_frame: Root = None

    def switch(self, name: str) -> None:
        new_frame = self.frame_classes[name]
        if self.current_frame is not None:
            self.current_frame.grid_forget()

        self.current_frame = new_frame
        self.current_frame.grid(row=0, column=0, sticky="nsew")

    def start_mainloop(self) -> None:
        self.root.mainloop()

    def stop_mainloop(self, *_) -> None:
        self.root.destroy()


import tkinter as tk
from tkinter import messagebox

from ..Views.MainView import MainView
from ..Views.HomeView import HomeView
from ...MayakoData import MayakoData
from ...Client import Client

class HomeController:

    main_view: MainView
    home_frame: HomeView

    def __init__(self, main_view: MainView, model: MayakoData, client: Client) -> None:
        self.model = model
        self.main_view = main_view
        self.client = client
        self.home_frame = self.main_view.frame_classes["home"]

        self._load_wifi_profiles_into_list()
        self._load_microcontrollers_into_list()
        self._view_bind()
        self._model_subscribe()

        for mc in self.model.get_microcontroller_capas():
            self.client.use_microcontroller(mc.identity)

    def _load_wifi_profiles_into_list(self) -> None:
        for key, profile in enumerate(self.model.get_wifi_profiles()):
            self.home_frame.wifi_listbox.insert(key, profile.format_string())

    def _load_microcontrollers_into_list(self) -> None:
        for key, microcontroller in enumerate(self.model.get_microcontroller_capas()):
            self.home_frame.devices_listbox.insert(key, microcontroller.format_string())

    def _view_bind(self) -> None:
        #devices
        self.home_frame.get_arduino_btn.config(command=self._change_frame_to_download_arduino)
        self.home_frame.add_device_btn.config(command=self._add_new_device)
        self.home_frame.remove_device_btn.config(command=self._remove_device)
        self.home_frame.device_details_btn.config(command=self._change_frame_to_device_details)

        #wifi profiles
        self.home_frame.add_wifi_btn.config(command=self._add_new_wifi_profile)
        self.home_frame.remove_wifi_btn.config(command=self._remove_wifi_profile)
        self.home_frame.wifi_details_btn.config(command=self._change_frame_to_wifi_profile_details)

    def _model_subscribe(self) -> None:
        self.model.subscribe("wifi_profiles_update", self._update_wifi_profiles)
        self.model.subscribe("microcontroller_capa_update", self._update_microcontrollers)

    def _update_wifi_profiles(self, _) -> None:
        self.home_frame.wifi_listbox.delete(0, tk.END)

        self._load_wifi_profiles_into_list()

    def _update_microcontrollers(self, _) -> None:
        self.home_frame.devices_listbox.delete(0, tk.END)

        self._load_microcontrollers_into_list()

    #download arduino frame
    def _change_frame_to_download_arduino(self) -> None:
        self.main_view.switch("download")

    #devices
    def _add_new_device(self) -> None:
        self.main_view.switch("add_device")

    def _remove_device(self) -> None:
        selected_item = self.home_frame.devices_listbox.curselection()

        if selected_item:
            item = self.model.get_microcontroller_capas()[selected_item[0]]
            self.model.remove_microcontroller_capa(item)

        else:
            messagebox.showerror(title="Error", message="no item selected")

    def _change_frame_to_device_details(self) -> None:
        selected_item = self.home_frame.devices_listbox.curselection()

        if selected_item:
            self.main_view.switch("device_details")
            self.model.select_microcontroller_capa(self.model.get_microcontroller_capas()[selected_item[0]])
    
        else:
            messagebox.showerror(title="Error", message="select an item to view")

    #wifi profiles
    def _add_new_wifi_profile(self) -> None:
        self.main_view.switch("add_wifi")

    def _remove_wifi_profile(self) -> None:
        selected_item = self.home_frame.wifi_listbox.curselection()

        if selected_item:
            item = self.model.get_wifi_profiles()[selected_item[0]]
            self.model.remove_wifi_profile(item)
        else:
            messagebox.showerror(title="Error", message="no item selected")

    def _change_frame_to_wifi_profile_details(self) -> None:
        selected_item = self.home_frame.wifi_listbox.curselection()

        if selected_item:
            self.main_view.switch("edit_wifi")
            self.model.select_wifi_profile(self.model.get_wifi_profiles()[selected_item[0]])
        else:
            messagebox.showerror(title="Error", message="select an item to view")


from dataclasses import dataclass, asdict
from typing import List, Dict, Self

@dataclass
class BLEInfo:
    address: str
    name: str
    service_uuid: str
    characteristic_uuid: str

    def to_dict(self) -> Dict:
        """makes a dictionary from the instance"""
        return asdict(self)
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """creates an instance of this class with information from a dictionary"""
        return cls(**data)


import os
import requests
from requests import Response
from typing import List, Tuple, Optional, Dict


class GithubRepositoryDownloader:
    """
    this class downloads specific folders and files from github repositories,
    stores them in the desired location on the local machine, and returns a status on the requested files.
    """

    def __init__(self, repository_url: str, destination_path: str, github_token: str) -> None:
        """
        initializes the downloader with the github repository URL and the destination path.

        Args
            repository_url (str):
                the url of a public github repository
                Example: https://github.com/vairasza/m5test
            destination_path (str);
                the destination folder in relation to the current working directory
                Example: arduino
        """
        self.repository_url = repository_url
        self.destination_path = destination_path
        self.github_token = github_token
        repo_info = self.repository_url.rstrip('/').split("/")
        self.owner = repo_info[-2]
        self.repository_name = repo_info[-1]

        self.api_url = f"https://api.github.com/repos/{self.owner}/{self.repository_name}/contents/"
        self.download_status: List[Tuple[str, str]] = []

        self._create_folder(self.destination_path)

    def _create_folder(self, folder_path: str) -> None:
        """
        creates the destination folder if it does not yet exist.
        
        Args
            folder_path (str):
                Example: lib
        """
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

    def _download_content(self, url: str) -> Optional[Response]:
        """
        downloads the content from the gtihub repository and returns the response object.
        The response can be a list (folder item) or a dict (file).
        The reponse object must be converted to json with response.json(). it results in a list of dictionaries or a dictionary. list ^= folder; dict ^= file

        Args
            path
                url for github content api

        Return
            Response or None:
                a request response object or none is the result of this function
        """

        if self.github_token:
            headers = { 'Authorization': f"Bearer {self.github_token}" }
            response = requests.get(url, headers=headers)
        else:
            response = requests.get(url)

        if response.status_code == 200:
            return response
        else:
            return None

    def _record_status(self, path: str, success: bool) -> None:
        """
        records the download status of a file or folder.
        
        Args
            path (str):
                Example: lib/test/Accelerometer.cpp 
            success (bool):
                if the content could be downloaded successfully or not
        """
        status = "success" if success else "fail"
        self.download_status.append((path, status))

    def _download_folder(self, dest_folder_path: str, folder_items: List[Dict]) -> None:
        """
        downloads the contents of a folder recursivly.

        Args
            dest_path (str):
                Example lib
            folder_items (List[Dict]):
                a list of dictionaries which describe the content of a github repo file. see source in top comment.
        """
        full_dest_path = os.path.join(self.destination_path, dest_folder_path)
        self._create_folder(full_dest_path)

        for folder_item in folder_items:
            item_path = folder_item['path']
            if folder_item['type'] == 'file':
                self._download_file(item_path, folder_item)

            elif folder_item['type'] == 'dir':
                url = self.api_url + item_path
                response = self._download_content(url)
                if not response:
                    self._record_status(item_path, False)
                    continue

                subfolder_info = response.json()
                self._download_folder(item_path, subfolder_info)

    def _download_file(self, dest_file_path: str, file_info: Dict) -> None:
        """
        downloads a single file and maintains the repositorys directory structure in the destination.

        Args
            dest_path (str):
                Example lib
            file_info (Dict):
                a dictionary which describe the content of a github repo file. see source in top comment.
        """
        file_url = file_info.get('download_url')

        if not file_url:
            self._record_status(dest_file_path, False)
            return

        #check if structure exists
        file_destination_dir = os.path.dirname(os.path.join(self.destination_path, dest_file_path))
        self._create_folder(file_destination_dir)

        try:
            file_response = self._download_content(file_url)
            file_destination = os.path.join(self.destination_path, dest_file_path)

            with open(file_destination, 'wb') as file:
                file.write(file_response.content)

            self._record_status(dest_file_path, True)

        except Exception as e:
            self._record_status(dest_file_path, False)

    def download(self, paths: List[str]) -> List[Tuple[str, str]]:
        """
        downloads only the specified files and folders
        
        Args
            paths (List[str]):
                Example: ["platformio.ini", "lib", "src"]
        
        Return
            List[Tuple[str, str]]
                Example  [(Success:,lib/README),(Fail:,lib/test/Accelerometer.cpp)]
        """
        for path in paths:
            url = self.api_url + path
            response = self._download_content(url)
            if not response:
                self._record_status(path, False)
                continue

            item_info = response.json()

            # ceck if item_info is a list => download_folder
            if isinstance(item_info, list):
                self._download_folder(path, item_info)
            # otherwise, it's a file
            else:
                self._download_file(path, item_info)

        return self.download_status

if __name__ == "__main__":
    #example usage
    repo_url = "https://github.com/vairasza/m5test"
    destination = "arduino"
    selected_paths = ["platformio.ini", "lib", "src"]
    g = GithubRepositoryDownloader(repository_url=repo_url, destination_path=destination)
    results = g.download(selected_paths)

    for path, status in results:
        print(f"{status}: {path}")



from dataclasses import dataclass, asdict
from typing import Dict, Self

@dataclass
class SerialInfo:
    """
    the is a model class representing serial information retrieved with mayako/Service/SerialScanner.py

    Attributes:
        port (str): the serial port on the target machine; Example: /dev/ttyUSB0
        serial_number (str): a 8 byte hex; may be empty
        description (str): describes the connected device e.g. information about the technology, the kind of device, etc.; may be empty
    """
    port: str
    serial_number: str
    description: str

    def format_string(self) -> str:
        """
        makes a string from the properties of the class.
        
        Returns:
            str: a visual representation of the class used in Views
        """
        return f"Port: {self.port} -- Serial Number: {self.serial_number}"
    
    def to_dict(self) -> Dict:
        """makes a dictionary from the class"""
        return asdict(self)
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """creates an instance of this class with information from a dictionary"""
        return cls(**data)


import tkinter as tk
from tkinter import messagebox

from ...Models.User import User
from ..Views.MainView import MainView
from ..Views.DownloadArduinoView import DownloadArduinoView
from ...Service.GithubRepositoryDownloader import GithubRepositoryDownloader
from ...MayakoData import MayakoData
from ...Config import CNETWORK
from ...Service.MacAdress import get_macaddress
from ...Utils.Logger import LoggerType, LoggerInstance
from ...Client import Client

class DownloadArduinoController:
    """
    this class downloads parts of a github repository into a selected folder.

    this is a controller class for the tkinter GUI.
    """
    logger: LoggerType
    main_view: MainView
    download_frame: DownloadArduinoView
    uploading: bool

    def __init__(self, main_view: MainView, model: MayakoData, client: Client) -> None:
        self.logger = LoggerInstance.get()
        self.main_view = main_view
        self.model = model
        self.client = client
        self.download_frame = self.main_view.frame_classes["download"]#type: ignore
        self._bind()
        self.uploading = False

    def _bind(self) -> None:
        self.download_frame.return_btn.config(command=self._return_to_main)
        self.download_frame.download_btn.config(command=self._dowload_repository)

    def _return_to_main(self) -> None:
        self.main_view.switch("home")
        self.download_frame.repo_url_entry.delete(0, tk.END)
        self.download_frame.destination_entry.delete(0, tk.END)
        self.download_frame.gh_token_entry.delete(0, tk.END)
        self.uploading = False
        self.download_frame.status_label.config(fg=self.download_frame.status_text[0][1])
        self.download_frame.status_label.config(text=self.download_frame.status_text[0][0])
        self.download_frame.result_text.delete("1.0", tk.END)

    def _dowload_repository(self) -> None:
        if self.uploading: return

        gh_url = self.download_frame.repo_url_entry.get()
        token = self.download_frame.gh_token_entry.get()
        dest = self.download_frame.destination_entry.get()

        if gh_url == "":
            messagebox.showerror(title="Error", message="Github Repostory URL must be a valid URL to a GitHub repository")

            return
        
        if dest == "":
            messagebox.showerror(title="Error", message="destination must be a valid folder")

            return

        self.uploading = True

        #show status in status label - progress
        self.show_progress_status()

        try:
            downloader = GithubRepositoryDownloader(repository_url=gh_url, destination_path=dest, github_token=token)
            results = downloader.download(CNETWORK.SELECTED_PATHS)

            for path, status in results:
                self.download_frame.result_text.insert(tk.END, f"{status}: {path}\n")

            user_mac_address = get_macaddress()
            user = User(user_mac_address=user_mac_address, arduino_folder=dest)
            self.model.add_user(user)

            self.show_success_status()
        
        except Exception as e:
            #show error status
            self.logger.error(e)
            self.show_error_status()

            messagebox.showerror(title="Error", message="there was an error while uploading")
        
        self.uploading = False

    def show_error_status(self):
        self.download_frame.status_label.config(text=self.download_frame.status_text[1][0])
        self.download_frame.status_label.config(fg=self.download_frame.status_text[1][1])

    def show_success_status(self):
        self.download_frame.status_label.config(text=self.download_frame.status_text[2][0])
        self.download_frame.status_label.config(fg=self.download_frame.status_text[2][1])

    def show_progress_status(self):
        self.download_frame.status_label.config(fg=self.download_frame.status_text[3][1])
        self.download_frame.status_label.config(text=self.download_frame.status_text[3][0])
        self.download_frame.status_label.update()


from bleak import BleakClient
import asyncio
from queue import Queue
import threading
from typing import Optional

from .NetworkInterface import NetworkInterface
from ..Config import CNETWORK, CPACKET
from .Packet import Packet
from ..Utils.Logger import LoggerInstance, LoggerType

class BLE(NetworkInterface):
    """
    Source
        https://www.instructables.com/Starting-and-Stopping-Python-Threads-With-Events-i/
        https://stackoverflow.com/a/68121144
        https://docs.python.org/3/library/asyncio-eventloop.html
        https://stackoverflow.com/a/71489745
    """
    _logger: LoggerType
    _address: str
    _service_uuid : str
    _characteristic_uuid: str
    _timeout: int
    _client: BleakClient
    _read_queue: Queue[Packet]
    _write_queue: Queue[Packet]
    _loop: asyncio.AbstractEventLoop
    _thread: threading.Thread
    _connected: bool
    _received_data: bytearray
    _expected_payload_size = Optional[int]

    def __init__(self, address: str, service_uuid: str, characteristic_uuid: str, timeout: int = CNETWORK.TIMEOUT) -> None:
        self._address = address
        self._service_uuid = service_uuid
        self._characteristic_uuid = characteristic_uuid
        self._timeout = timeout
        self._logger = LoggerInstance.get()
        self._client = BleakClient(self._address, timeout=self._timeout)
        self._read_queue = Queue()
        self._write_queue = Queue()
        self._loop = asyncio.new_event_loop()
        self._thread = None
        self._connected = False
        self._received_data = bytearray()
        self._expected_payload_size = None

    def connect(self) -> None:
        if self._thread and self._thread.is_alive():
            self._logger.debug("BLE: thread was still alive")
            self._thread.join()

        self._thread = threading.Thread(target=self._start_event_loop)
        self._thread.start()
    
    def _start_event_loop(self):
        # Create a new event loop for this thread
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._loop.run_until_complete(self._handle_device())

    async def _handle_device(self) -> None:
        try:
            await self._connect_client()

            if self._connected:
                notify_task = self._loop.create_task(self._start_notify())
                write_task = self._loop.create_task(self._write_data())

                await asyncio.gather(notify_task, write_task)

        except asyncio.CancelledError:
            self._logger.error(f"BLE: stop notify failed: asyncio.CancelledError")
        finally:
            await self._stop_notify()
            await self._client.disconnect()

    async def _connect_client(self) -> None:
        try:
            await self._client.connect()
            await self._client.get_services()
            self._connected = self._client.is_connected
            self._logger.debug("BLE: connected")
        except Exception as e:
            self._logger.error(f"BLE: connecting client failed: {e}")
            self._connected = False
            
    async def _start_notify(self):
        try:
            await self._client.start_notify(self._characteristic_uuid, self._notification_handler)
        except Exception as e:
            self._logger.error(f"BLE: start notify failed: {e}")

    async def _stop_notify(self):
        try:
            await self._client.stop_notify(self._characteristic_uuid)
        except Exception as e:
            self._logger.error(f"BLE: stop notify failed: {e}")
    
    def _notification_handler(self, sender: str, data: bytes):
        """Handles incoming BLE data and defragments it if needed."""
        # Append current data to accumulated packet
        self._received_data.extend(data)

        # If we don't know the expected payload size yet, calculate it from the header
        if self._expected_payload_size is None and len(self._received_data) >= CPACKET.HEADER_SIZE:
            # Assuming the payload size is stored in the header at bytes 8-10
            self._expected_payload_size = int.from_bytes(self._received_data[8:10], byteorder='big')
            #self._logger.debug(f"Expected payload size: {self._expected_payload_size}")

        # Check if we've received the full packet
        if self._expected_payload_size and len(self._received_data) >= self._expected_payload_size:
            # Full packet has been received, process it
            self._process_complete_packet(self._received_data)
            self._received_data.clear()  # Clear buffer for next packet
            self._expected_payload_size = None  # Reset for the next packet

    def _process_complete_packet(self, packet: bytearray):
        """Processes the complete packet after defragmentation."""
        #self._logger.debug("Complete packet received:")
        if len(packet) < CPACKET.HEADER_SIZE:
            self._logger.debug("BLE: packet too small to be valid")
            return
        
        try:
            # Deserialize packet header and payload
            header_data = packet[:CPACKET.HEADER_SIZE]
            payload_data = packet[CPACKET.HEADER_SIZE:]

            pkt = Packet()
            pkt.deserialize_header(header_data)
            pkt.deserialize_payload(payload_data)

            # Push the fully received and processed packet into the read queue

            self._read_queue.put(pkt)
            
        except Exception as e:
            self._logger.error(f"BLE: Error processing packet: {e}")

    async def _write_data(self) -> None:
        while self._connected:
            if not self._write_queue.empty():
                packet = self._write_queue.get()
                serialised_data = packet.serialize()

                try:
                    await self._client.write_gatt_char(self._characteristic_uuid, serialised_data)
                except Exception as e:
                    self._logger.error(f"BLE: failed writing data: {e}")
                    pass

            await asyncio.sleep(CNETWORK.BUSY_WAITING_TIMEOUT)

    def disconnect(self) -> None:
        if self._connected:
            self._connected = False  # Set connected status to False
            # Gracefully stop the BLE connection in the asyncio loop
            asyncio.run_coroutine_threadsafe(self._disconnect_client(), self._loop)

        if self._thread and self._thread.is_alive():
            self._thread.join()

        if self._loop.is_running():
            self._loop.run_until_complete(self._loop.shutdown_asyncgens())
            self._loop.stop()

        self._loop.close()
        self._logger.debug("BLE: network connection was closed")

    async def _disconnect_client(self) -> None:
        try:
            await self._stop_notify()
            await self._client.disconnect()
        except Exception as e:
            self._logger.error(e)

    def write(self, data: Packet) -> None:
        self._write_queue.put(data)

    def read(self) -> Optional[Packet]:
        if self._read_queue.qsize() > 0:            
            return self._read_queue.get()

        return None

    def check_connection(self) -> bool:
        return self._connected
    
    def get_protocol_name(self) -> str:
        return CNETWORK.BLE_NAME


from dataclasses import dataclass, asdict, field
from typing import Self, Dict, List, Optional
from ..Network.NetworkInterface import NetworkProtocols
from ..Models.SensorCapabilities import SensorCapabilities
from ..Models.ActuatorCapabilities import ActuatorCapabilities

@dataclass
class MicroControllerCapabilities:
    """
    Sources:
        https://stackoverflow.com/a/51976841
    """
    identity: str

    #these values influence the record
    include_timestamp: bool = False
    include_sequence: bool = False
    delay: int = 0
    duration: int = 0
    max_samples: int = 0

    #status
    online: bool = False
    wifi_key: Optional[str] = None  #this is only the wifi_key to the associated wifi_key
    serial_port: Optional[str] = None
    ble_address: Optional[str] = None
    battery_percentage: int = 0
    battery_charging: bool = False
    sensors: List[SensorCapabilities] = field(default_factory=list)
    actuators: List[ActuatorCapabilities] = field(default_factory=list)
    protocol: NetworkProtocols = field(default=NetworkProtocols.BLE)
    name: str = ""

    def format_string(self) -> str:
        """returns a string with a representation of the instance for Views"""
        return f"identity: {self.identity} -- protocol: {self.protocol.name} -- online: {self.online} -- serial port: {self.serial_port}"

    def to_dict(self) -> Dict:
        """makes a dictionary from the instance"""
        return asdict(self)
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """
        makes an instance of the according class using a dictionary.

        not existing values are set to default values. identity must exist as it is an integral part.

        Args:
            data (Dict): dictionary with instance properties

        Returns:
            Self: an instance of the class with the assigned values
        """
        default_data = {
            'include_timestamp': False,
            'include_sequence': False,
            'delay': 0,
            'duration': 0,
            'max_samples': 0,
            'online': False,
            'protocol': NetworkProtocols.BLE,
            'wifi_key': None,
            'serial_port': None,
            'ble_address': None,
            'battery_percentage': 0,
            'battery_charging': False,
            'name': "",
            'sensors': [],
            'actuators': []
            #identity must be provided
        }
        
        merged_data = {**default_data, **data}

        #ensure that proptocol is converted from string to enum
        if isinstance(merged_data['protocol'], str):
            merged_data['protocol'] = NetworkProtocols[merged_data['protocol']]

        merged_data['sensors'] = [
            SensorCapabilities(**sensor) if isinstance(sensor, dict) else sensor
            for sensor in merged_data['sensors']
        ]

        # Convert actuators from list of dicts to list of ActuatorCapabilities
        merged_data['actuators'] = [
            ActuatorCapabilities(**actuator) if isinstance(actuator, dict) else actuator
            for actuator in merged_data['actuators']
        ]
            
        return cls(**merged_data)


from tkinter import Frame, Label, Button, Listbox, Scrollbar, VERTICAL, SINGLE
import tkinter as tk

class HomeView(Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Configure grid
        self.grid_columnconfigure(0, weight=0)
        self.grid_columnconfigure(1, weight=1)  # No weight for the scrollbar column

        # Download Arduino Project from Github
        self.get_arduino_btn = Button(self, text="Get Arduino Project")
        self.get_arduino_btn.grid(row=0, column=0, columnspan=3, padx=10, pady=10, sticky="we")

        # Devices Section
        self.devices_label = Label(self, text="Devices")
        self.devices_label.grid(row=1, column=0, padx=10, pady=10, sticky="ew")

        self.devices_listbox = Listbox(self, selectmode=SINGLE)
        self.devices_listbox.grid(row=2, column=0, columnspan=3, padx=(10, 0), pady=5, sticky="nsew")

        self.devices_scrollbar = Scrollbar(self, orient=VERTICAL, command=self.devices_listbox.yview)
        self.devices_scrollbar.grid(row=2, column=3, padx=(0, 10), pady=5, sticky="ns")
        self.devices_listbox.config(yscrollcommand=self.devices_scrollbar.set)

        self.add_device_btn = Button(self, text="Add")
        self.add_device_btn.grid(row=3, column=0, padx=10, pady=5)

        self.remove_device_btn = Button(self, text="Remove")
        self.remove_device_btn.grid(row=3, column=1, padx=10, pady=5)

        self.device_details_btn = Button(self, text="Details")
        self.device_details_btn.grid(row=3, column=2, padx=10, pady=5)

        # Wi-Fi Profiles Section
        self.wifi_label = Label(self, text="Wi-Fi Profiles")
        self.wifi_label.grid(row=4, column=0, padx=10, pady=10, sticky="ew")

        self.wifi_listbox = Listbox(self, selectmode=SINGLE)
        self.wifi_listbox.grid(row=5, column=0, columnspan=3, padx=(10, 0), pady=5, sticky="nsew")

        self.wifi_scrollbar = Scrollbar(self, orient=VERTICAL, command=self.wifi_listbox.yview)
        self.wifi_scrollbar.grid(row=5, column=3, padx=(0, 10), pady=5, sticky="ns")
        self.wifi_listbox.config(yscrollcommand=self.wifi_scrollbar.set)

        self.add_wifi_btn = Button(self, text="Add")
        self.add_wifi_btn.grid(row=6, column=0, padx=10, pady=5)

        self.remove_wifi_btn = Button(self, text="Remove")
        self.remove_wifi_btn.grid(row=6, column=1, padx=10, pady=5)

        self.wifi_details_btn = Button(self, text="Edit")
        self.wifi_details_btn.grid(row=6, column=2, padx=10, pady=5)

if __name__ == "__main__":
    root = tk.Tk()
    root.title("Add New Device")
    root.geometry("400x620")
    add_device_view = HomeView(root)
    add_device_view.pack(fill="both", expand=True)
    root.mainloop()



import re
from .Exceptions import *
from ..Config import CERROR_MESSAGES

class Identity:

    '''Identity manages identity keys to uniquely distinguish devices and other things.
    
    End Users do not explicitly use the Identity class, it is part of the devices

    Attributes:
        ACC_01 (str): An example for a class attribute which can be used as a identity key to name Sensors, Actuators, Microcontrollers or WiFiProfiles
        TEMP_01 (str): A second example
        WP_01 (str): Another example
    '''

    def __init__(self) -> None:
        '''Identity should not be instantiated to avoid double keys.'''
        raise NotImplementedError(CERROR_MESSAGES.CLASS_NO_INIT)

    #https://stackoverflow.com/a/71546645
    @classmethod
    def __str__(cls) -> None:
        '''Makes a report on the attributes used in the class.'''
        xs = []
        for name, value in vars(cls).items():
            if not (name.startswith('__') or isinstance(value, classmethod)):
                xs.append(value)
        return xs

    @classmethod
    def register(cls, identity: str) -> None:
        '''
        This function registers identity keys and throws if they are not unique.
        
        Identity keys are meant to be short to keep the syntax short.
        Identity keys are not meant for long description, in this case use the description attribute on the according classes.

        Identitiy may have 2 to 20 charactes that are a to z or numbers.

        Raises a custom exception if the attribute is already used.
        Raises a custom exception if the provided identity is empty or does not match the pattern.

        Args:
            identity (str): An identity key chosen by the end user to uniquely identify devices in the framework.
        '''
        if type(identity) is not str or identity == "":
            raise IdentitiyPatternError(identity)
        
        pattern = r'^[A-Za-z0-9_]{2,20}$'

        if not re.search(pattern, identity):
            raise IdentitiyPatternError(identity)

        if hasattr(cls, identity):
            raise IdentitiyKeyUsedError(identity)
        
        setattr(cls, identity, identity)


from enum import Enum
import logging
from logging.handlers import QueueHandler
from queue import Queue
import sys

class LoggingLevel(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARN
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class LoggingDestination(Enum):
    STDOUT = "STDOUT"
    QUEUE = "QUEUE"

LoggerType = logging.Logger

class Logger:
    # IMPORTANT: use LoggingLevel.INFO if you want to see events from the network such as BATTERY_INFO
    #instantiate this class in client with enable/disable logger => enable is sufficient i guess
    #only for loop events
    #for setup use execptions
    #https://stackoverflow.com/a/44401529
    _LOG_FORMAT:str = '%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s'
    #TODO: debug formater w
    _logger: LoggerType

    def init_logger(self, logging_level: LoggingLevel = LoggingLevel.INFO, destination: LoggingDestination = LoggingDestination.STDOUT, output_format: str = _LOG_FORMAT, queue: Queue = None) -> None:
        self._logging_level: LoggingLevel = logging_level
        self._destination = destination
        self._output_format = output_format
        self._queue = queue
        
        """ if self._logger:
            raise Exception("Logger instance already exists") """

        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(LoggingLevel.DEBUG.name)

        self._handler = self._define_handler(self._destination, self._queue)
        self._handler.setLevel(self._logging_level.value)
        fmt = logging.Formatter(self._output_format)
        self._handler.setFormatter(fmt)

        self._logger.addHandler(self._handler)

    #need a place where we write the logs: file, std:out, network, whatever
    def _define_handler(self, destination: LoggingDestination, queue: Queue) -> logging.Handler:
        output: logging.Handler
        
        if destination == LoggingDestination.STDOUT:
            output = logging.StreamHandler(stream=sys.stdout)

        elif destination == LoggingDestination.QUEUE:
            if type(queue) is not Queue:
                raise TypeError(f"using {LoggingDestination.QUEUE} requires a queue")
            
            output = QueueHandler(queue)

        else:
            raise NotImplementedError(f"{destination.value} is not implemented")
        
        return output
    
    def get(self) -> LoggerType:
        return self._logger

LoggerInstance = Logger()


import struct
from crc import Calculator, Configuration

from ..Config import CPACKET, CERROR_MESSAGES
from ..Utils.Logger import LoggerType, LoggerInstance

class Packet:
    """
    This module deserialises binary data received with an implemented protocol while also serialising data for network transfer. It provides functionality to split the header from the payload, performs checksum test and also verifies if a byte contains the flag for a mayakoProtocol packet.

    The structure of the binary packet is described in protocol_interface.md

    Do not access the members of the class directly, use the getter and setter functions to avoid unexpected behaviour.

    Sources:
        https://docs.python.org/3/library/struct.html
        https://pypi.org/project/crc/

    Attributes:
        _method : int (8bit)
            makes the first byte of each packet and the header. indicates the purpose of the packet in the mayako framework. can be one of 6 hexademical values which can be found in config.py
        _node_identity : int (32 bit)
            contains the name of the microcontroller or the client application which sent the packet. Node identity is a fixed-size 4 byte ASCII integer so that it can be integrated in the binary header. This lets everyone directly recognise the source of the packet and requires no further identification inside the payload. The node identity for each microcontroller and client application must be unique throughout the project and (probably) also in BLE/WiFi range. The node identity for the microcontroller is defined with the upload of a new build and can not changed for commands. 
        _sequence : int (16 bit)
            there is a difference between _sequence in the packet and sequence in the payload of sensor data. _sequence works across generic packets and tracks the order of the  packets. It is intended to work with ACK packets. the microcontroller app uses a 16 bit unsigned integer for _sequence, therefore we must be cautious and can only use _sequence until 2¹⁶-1.
        _checksum : int (8 bit)
            we use CRC to calculate the checksum of the payload to ensure the integrity of the data. corrupted or incomplete data fails to test the checksum which should trigger a ACK packet with a retry request. the parameters for the CRC calculation can be found in protocol_interface.md.
        _payload_size : int (16 bit)
            to ensure that we can savely read the content of each packet, besides starting with the method flag, we integrated a payload size 2 byte field in the header on position 9 and 10. this contains the length of the payload and helps the protocols to read data until the end. therefore we do not have to rely on a newline character and can read/write binary data.
        _payload : bytes
            payload is of variable size and is encoded to utf-8 using JSON format. the size is sent in the header. more on the structure of payload can be read in protocol_interface.md.
            NYI: currently only text encoding is available. change encoding based on data (video data, audio data, etc.). we recommend an additional header field which indicates that encoding that has to be updated in the network protocol specification.
        _header_size : int (10 byte)
            the header size is an integral part of the packet and has a fix size of 10 bytes. this must be uniformly enforced across all implementation. failure to comply may result in unexpected behaviour, communication errors and data misinterpretations. Following th header, the payload has variable size only limited by the MTU of the implemented protocol. the header size is not integrated into the header because of its constant nature.
        _packet_size: int (variable)
            the sum of _header_size and _payload_size

    Todo:
        * encoding field in the header
    """
    _logger: LoggerType
    _method: int
    _node_identity: int
    _sequence: int
    _checksum: int
    _payload_size: int
    _payload: bytes
    _header_size: int
    _packet_size: int

    def __init__(self) -> None:
        """
        set all members to None. incorrect usage results in members still being None.
        """
        self._logger = LoggerInstance.get()
        self._method = None
        self._node_identity = None
        self._sequence = 0
        self._checksum = None
        self._payload_size = None
        self._payload = None
        self._header_size = CPACKET.HEADER_SIZE
        self._packet_size = None

    def set_method(self, method: int) -> None:
        """
        sets _method. best use the constants from Config.py as argument.
        
        Args:
            method (int): the argument must be a 1 byte hex value

        Returns:
            None
        """

        self._method = method

    def set_node_identity(self, node_identity: str) -> None:
        """
        node_identity is a 4 byte long name. this function does not validate if it is a 4 byte ASCII string. node identity is passed as a string that is then converted to bytes with ascii encoding.

        Args:
            node_identity (str): 4 byte string that can be encoded to ASCII

        Returns:
            None
        """
        if len(node_identity) != CPACKET.NODE_IDENTITY_LENGTH:
            raise ValueError(CERROR_MESSAGES.NODE_IDENTITY_LENGTH)

        self._node_identity = int.from_bytes(node_identity.encode(CPACKET.HEADER_ENCODING))

    def _set_sequence(self, sequence: int) -> None:
        """
        sequence is a 2 byte number. it is automatically managed by the Integrity Manager is may not be set manually otherwise resulting in unexpected behaviour and maybe congestion issues.

        Args:
            sequence (int): sequence indicates the order of packets on the network

        Returns:
            None
        """
        self._sequence = sequence

    def _set_checksum(self, checksum: int) -> None:
        """
        checksum is a 8 bit number. it is automatically managed. do not set this manually otherwise unexpected behaviour may occur.

        Args:
            checksum (int): 8 bit CRC checksum

        Returns:
            None
        """
        self._checksum = checksum

    def _set_payload_size(self, payload_size: int) -> None:
        """
        payload_size is an 16 bit number. it is automatically managed. do not set this manually otherwise unexpected behaviour may occur.

        Args:
            payload_size (int): size of the payload

        Returns:
            None
        """
        self._payload_size = payload_size

    def set_payload(self, payload: str) -> None:
        """
        the payload is first encoded to uft-8. We calculate the payloadsize and the checksum from it.

        Args:
            payload (str): a JSON string with commands, data, or error messages

        Returns:
            None
        """
        self._payload = payload.encode(CPACKET.PAYLOAD_ENCODING)
        self._payload_size  = len(self._payload)
        self._checksum = self.calculate_checksum(self._payload)

    def get_method(self) -> int:
        """
        Returns the method.

        Args:
            None

        Returns:
            _method (int) 8 bit integer
        """
        return self._method
    
    def get_node_identity(self) -> str:
        """
        Returns the node identity.

        Args:
            None

        Returns:
            _node_identity (str) identity of the microcontroller or client application which sent the packet
        """
        return self._node_identity.to_bytes(4).decode(CPACKET.HEADER_ENCODING)
    
    def get_sequence(self) -> int:
        """
        Returns the sequence.

        Args:
            None

        Returns:
            _sequence (int)
        """
        return self._sequence
    
    def get_checksum(self) -> int:
        """
        Returns the checksum
        
        Args:
            None

        Returns:
            _checksum (int) 8 bit CRC checksum
        """
        return self._checksum
    
    def get_payload_size(self) -> int:
        """
        Returns the payload size

        Args:
            None

        Returns:
            _payload_size (int)
        """
        return self._payload_size
    
    def get_payload(self) -> str:
        """
        Returns the payload.

        Args:
            None

        Returns:
            _payload (str)
        """
        
        return self._payload.decode(CPACKET.PAYLOAD_ENCODING)
    
    def get_packet_size(self) -> int:
        """
        Returns the packet size.

        Args:
            None

        Returns:
            packet size (int)
        """
        return self._packet_size
    
    def serialize(self) -> bytes:
        """
        Before using serialise method, node_identity, and payload must have been set. Raises an exception if one or more are None. packet size is calculated from header and payload size. this is used to compare with the MTU of the implemented procotol before sending over the network, probably encountering errors or corrupted data. the header is built with the package struct. the 10 byte header is formated with format characters that can be found in config.py. it packs method, node_identity, sequence, checksum and payload. it results in a byte array. header is then combined with the byte array of the payload to complete a ready to send packet.

        Source:
            #https://docs.python.org/3/library/struct.html

        Args:
            None

        Returns:
            packet (bytes): the packet contains header and payload an can be sent over the network like this

        Raises:
            ValueError: if _method, _node_identity, or _payload are not set (None) before calling seriliase.

        """
        if self._method == None or self._node_identity == None or self._payload == None:
            raise ValueError(CERROR_MESSAGES.PACKET_NOT_SET)


        self._packet_size = self._header_size +  self._payload_size
        
        header = struct.pack(CPACKET.HEADER_FORMAT_CHARACTERS, self._method, self._node_identity, self._sequence, self._checksum, self._payload_size)

        return header + self._payload
    
    def deserialize_header(self, buffer: bytes) -> bool:
        """
        the deserialise method is split into two functions. this is due to the fact how we are reading the data in the implemented protocol read function. first we check if the buffer has enough data for the header, deserialise it, and result with a payload size. because we are reading binary data and do not rely on a stop character such as new line, we require the payload size from byte 9 and 10 before being able to read the payload.

        in this function we first unpack the 10 byte bytes array with the struct packet with a defined character format. this results in a tuple (length 5) which contains method, node_identity, sequence, checksum and payload_size. these are not returned but stored in the class as properties which should be later read with getter methods as needed. do use both deserialise methods before using getter methods.

        Args:
            buffer (bytes): the byte buffer read from the network starting with the method flag

        Returns:
            anonymous (boolean): if buffer is not HEADER_SIZE it returns false
        """
        if len(buffer) == CPACKET.HEADER_SIZE - 1:
            return False
        
        result: tuple = struct.unpack(CPACKET.HEADER_FORMAT_CHARACTERS, buffer)
        
        self._method = result[0]
        self._node_identity = result[1]
        self._sequence = result[2]
        self._checksum = result[3]
        self._payload_size = result[4]

        return True
    
    def deserialize_payload(self, buffer: bytes) -> bool:
        """
        deserialises the payload. this sets the payload as a byte array. use getter method to get the decoded paylaod.

        Args:
            buffer (bytes): the payload of a packet
        
        Returns:
            anonymous (boolean): if the buffer is not the payload_size it returns false
        """
        if len(buffer) < self._payload_size:
            self._logger.debug(len(buffer))
            self._logger.debug(self._payload_size)
            return False
        
        self._payload = buffer

        return True

    @staticmethod
    def calculate_checksum(payload: bytes) -> int:
        """
        static method to calcuate the CRC checksum

        Args:
            payload (bytes)

        Returns:
            checksum (int): 8 bit CRC checksum
        """
        crc_config = Configuration(width=CPACKET.CRC_WIDTH,
                                   polynomial=CPACKET.CRC_POLYNOMIAL,
                                   init_value=CPACKET.CRC_INIT_VALUE,
                                   final_xor_value=CPACKET.CRC_FINAL_XOR_VALUE, reverse_input=CPACKET.CRC_REVERSE_INPUT,
                                   reverse_output=CPACKET.CRC_REVERSE_OUTPUT)

        calculator = Calculator(configuration=crc_config)

        return calculator.checksum(payload)
    
    def verify_checksum(self) -> bool:
        """
        verifies the checksum with the properties of packet. used after deserialising the packet.

        Args:
            None

        Returns:
            anonymous (boolean): if the checksum of the checksum from the header and performing the checksum on the payload inplace match
        """
        if self._payload == None or self._checksum == None:
            raise ValueError(CERROR_MESSAGES.PACKET_NOT_SET)

        return self.calculate_checksum(self._payload) == self._checksum
    
    @staticmethod
    def verify_flag(flag: bytes) -> bool:
        """
        static method to verify a one byte integer. compares with the method flags of the mayakoProtocol. if it matches, this indicates that the passed flag is the first byte of a packet from this protocol. the flag may not be an integer so we try to convert it.

        Args:
            flag (bytes): one byte of data

        Returns:
            anonymous (boolean): if the flag is from the mayakoProtocol packet
        """
        if type(flag) != int:
            flag = int.from_bytes(flag)

        if flag == CPACKET.METHOD_ACKNOWLEDGEMENT:
            return True
        elif flag == CPACKET.METHOD_DATA:
            return True
        elif flag == CPACKET.METHOD_COMMAND:
            return True
        elif flag == CPACKET.METHOD_HEARTBEAT:
            return True
        elif flag == CPACKET.METHOD_DEBUG:
            return True
        elif flag == CPACKET.METHOD_INFO:
            return True
        elif flag == CPACKET.METHOD_ERROR:
            return True
        else:
            return False
        
    def verify_good_packet(self) -> bool:
        """
        this method is a combination of verify flag and perform checksum. requires a packet to be deserialised. Do not call before deserialising.

        Args:
            None

        Returns: 
            anonymous (boolean) if verify flag and checksum passed
        """
        
        return self.verify_flag(self.get_method()) and self.verify_checksum()


from .Device import Device
from ..MayakoData import MayakoData
from ..Models.ActuatorCapabilities import ActuatorCapabilities
from ..Command.CustomCommands import CustomCommands
from ..Config import CCLIENT

class Actuator(Device):

    def __init__(self, identity: str, mc_identity: str, model: MayakoData, commands: CustomCommands) -> None:
       super().__init__(identity=identity, mc_identity=mc_identity, model=model)
       self._commands = commands
       self._check_capabilities(self._identity)

    def define_capabilities(self, enable: bool = True) -> None:
        """
        this function lets the user change capabilties of the actuator subclass

        Args:
            enable (bool): if the actuator is enabled or not
        """
        old_capas = self._check_capabilities(actuator_identity=self._identity)

        new_capas = ActuatorCapabilities(identity=self._identity, enable=enable)

        if not self._compare_actuator_capabilities(old_capas=old_capas, new_capas=new_capas):
            self._logger.debug("updating actuator capability")
            self._model.update_actuator_capability(self._mc_identity, new_capas)
        else:
            self._logger.debug("no update on actuator capability")

    def _check_capabilities(self, actuator_identity: str) -> ActuatorCapabilities:
        """checks if there is a capability associated with the acutator identity"""
        capas = self._model.get_actuator_capability_by_identity(mc_identity=self._mc_identity, actuator_identity=actuator_identity)
        if not capas:
            raise Exception(f"the actuator with the provided identity {actuator_identity} does not exist in {CCLIENT.CONFIG_FILE_NAME}.")
        
        return capas

    def get_capabilities(self) -> ActuatorCapabilities:
        """returns the capabilities of the actuator"""
        return self._model.get_actuator_capability_by_identity(self._identity)

    def _compare_actuator_capabilities(self, old_capas: ActuatorCapabilities, new_capas: ActuatorCapabilities) -> bool:
        """compares the actuator capabilities"""
        return (
            old_capas.enable == new_capas.enable
        )


from abc import ABC, abstractmethod
from enum import Enum
from ..Network.Packet import Packet
from typing import Optional

class NetworkProtocols(str, Enum):
    """
    Indicates the Network Protocol in MicrocontrollerCapabilities
    
    Sources:
        https://stackoverflow.com/a/51976841
    """
    WIFI="WIFI"
    BLE="BLE"
    SERIAL="SERIAL"

class NetworkInterface(ABC):

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def connect(self) -> None:
        pass

    @abstractmethod
    def disconnect(self) -> None:
        pass

    @abstractmethod
    def write(self, data: Packet) -> None:
        pass

    @abstractmethod
    def read(self) -> Optional[Packet]:
        pass

    @abstractmethod
    def check_connection(self) -> bool:
        pass
    
    @abstractmethod
    def get_protocol_name(self) -> str:
        pass


from ..MayakoData import MayakoData
from ..Command.CustomCommands import CustomCommands
from .Actuator import Actuator

class LEDActuator(Actuator):

    def __init__(self, identity: str, mc_identity: str, model: MayakoData, commands: CustomCommands) -> None:
        super().__init__(identity, mc_identity, model, commands)

    def switch_on(self, index: int, red: int, green: int, blue: int) -> None:
        parameters = {
            "index": index,
            "red": red,
            "green": green,
            "blue": blue
        }
        self._commands.execute(self._mc_identity, "SWITCH_ON", parameters)

    def switch_off(self) -> None:
        self._commands.execute(self._mc_identity, "SWITCH_OFF")


from serial.tools.list_ports import comports
from typing import List
from ..Models.SerialInfo import SerialInfo

def scan_serial_ports(device_names: List[str]) -> List[SerialInfo]:
    """
    This function scans all serial ports on the machine (WIN, MACOS, LINUX). Arduino boards often have a description containing the name of the USB-to-UART bridge controller. Therefore we can filter the scanned ports by looking for keywords defined in Config.h > CNETWORK > SERIAL_ARDUINO_WORDS. We stored following information in a dataclass: the port of the found arduino board, the serial number to identify the board in another iteration, the description which contains the keywords.

    Source:
        https://github.com/pyserial/pyserial

    Args:
        device_names (List[str]):
            a list of words that are typical for the arduino serial port description

    Returns:
        list of ports which match our criteria for an arduino board
    """
    available_ports = comports()
    arduino_ports: List[SerialInfo] = []

    for port in available_ports:
        if len(device_names) != 0 and any(keyword in port.description for keyword in device_names):
            serial_number = getattr(port, "serial_number", "unknown")

            arduino_ports.append(SerialInfo(port.device, serial_number, port.description))

    return arduino_ports

if __name__ == "__main__":
    device_name = ["CP2104", "Arduino"]
    serial_ports = scan_serial_ports(device_name)


import tkinter as tk
from tkinter import messagebox
import json
from ..Views.MainView import MainView
from ..Views.DeviceDetailsView import DeviceDetailsView
from ...MayakoData import MayakoData
from ...Client import Client

#TODO
from ...Models.Button import Button
from ...Models.Temperature import Temperature
from ...Models.SensorCapabilities import SensorCapabilities
from ...Models.ActuatorCapabilities import ActuatorCapabilities

class DeviceDetailsController:

    main_view: MainView
    device_details_frame: DeviceDetailsView

    def __init__(self, main_view: MainView, model: MayakoData, client: Client) -> None:
        self.main_view = main_view
        self.model = model
        self.client = client
        self.device_details_frame = self.main_view.frame_classes["device_details"]
        self._bind()
        self._model_subscribe()

    def _bind(self) -> None:
        self.device_details_frame.return_btn.config(command=self._return_to_main)
        self.device_details_frame.update_capabilities_btn.config(command=self._update_capabilities)
        #self.device_details_frame.identify_device_btn.config(command=self._identify)

    def _return_to_main(self) -> None:
        self.main_view.switch("home")

    def _model_subscribe(self) -> None:
        self.model.subscribe("microcontroller_capa_selected", self._load_microcontroller)

    def _load_microcontroller(self, mayako_data: MayakoData) -> None:
        mc = mayako_data.selected_mc_capability
        self.device_details_frame.identity_text.config(text=mc.identity)
        self.device_details_frame.protocol_text.config(text=f"{mc.protocol.name} -- {"online" if mc.online else "offline"}")
        self.device_details_frame.battery_text.config(text=f"{mc.battery_percentage}% -- {'charging' if mc.battery_charging else 'running on battery'}")

        self.device_details_frame.capabilities_text.delete("1.0", tk.END)
        self.device_details_frame.capabilities_text.insert(tk.END, json.dumps(mc.to_dict(), indent=4))
        
    def _update_capabilities(self) -> None:
        mc_capa = self.model.selected_mc_capability
        #check if online => enable button
        
        try:
            #if mc_capa.identity not in self.client._microcontrollers:
            #    self.client.use_microcontroller(mc_identity=mc_capa.identity)
            self.client._commands.record_read(mc_identity=mc_capa.identity)

        except Exception as e:
            print(e)
            messagebox.showerror("ERROR", "device not available")

    """ def _identify(self) -> None:
        mc_capa = self.model.selected_mc_capability
        try:
            #if mc_capa.identity not in self.client._microcontrollers:
            #    self.client.use_microcontroller(mc_identity=mc_capa.identity)

            #self.client._commands.identify(mc_identity=mc_capa.identity, identity=mc_capa.identity)
            messagebox.showinfo("INFO", "this feature is not implemented.")
        except:
            messagebox.showerror("ERROR", "device not available") """


from tkinter import Frame, Label, Button, Listbox, Scrollbar, Entry, Checkbutton, BooleanVar
import tkinter as tk

class AddDeviceView(Frame):
    def __init__(self, *args, **kwargs):
        """
        Sources:
            https://stackoverflow.com/a/26366001
        """
        super().__init__(*args, **kwargs)

        self.grid_columnconfigure(0, weight=0)
        self.grid_columnconfigure(1, weight=1)

        #headline
        self.headline = Label(self, text="Add New Device")
        self.headline.grid(row=0, column=0, columnspan=2, padx=10, pady=10, sticky="w")

        #return button
        self.return_btn = Button(self, text="Return")
        self.return_btn.grid(row=0, column=1, padx=10, pady=10, sticky="e")

        #checkboxes
        self.ble_var = BooleanVar()
        self.wifi_var = BooleanVar()
        self.ble_check = Checkbutton(self, text="BLE", variable=self.ble_var, onvalue=True, offvalue=False)
        self.ble_check.grid(row=2, column=0, padx=10, pady=5, sticky="w")
        self.wifi_check = Checkbutton(self, text="WiFi", variable=self.wifi_var, onvalue=True, offvalue=False)
        self.wifi_check.grid(row=2, column=1, padx=10, pady=5, sticky="w")

        #list of wifi profiles
        self.wifi_profile_listbox = Listbox(self, selectmode="single", exportselection=False)
        #self.wifi_profile_listbox.grid(row=3, column=0, columnspan=2, padx=10, pady=5, sticky="nsew")
        self.wifi_profile_scrollbar = Scrollbar(self, orient="vertical", command=self.wifi_profile_listbox.yview)
        #self.wifi_profile_scrollbar.grid(row=3, column=1, padx=(0, 10), pady=5, sticky="nes")
        self.wifi_profile_listbox.config(yscrollcommand=self.wifi_profile_scrollbar.set)

        #identity text box
        self.identity_label = Label(self, text="Identity: (must be 4 chars long)")
        self.identity_label.grid(row=4, column=0, padx=10, pady=5, sticky="w")
        self.identity_entry = Entry(self)
        self.identity_entry.grid(row=4, column=1, padx=10, pady=5, sticky="we")

        #select devices text
        self.select_devices_label = Label(self, text="Select the serial port the device is connected to:")
        self.select_devices_label.grid(row=5, column=0, padx=10, pady=5, sticky="w")
        self.serial_devices_update_btn = Button(self, text="Update")
        self.serial_devices_update_btn.grid(row=5, column=1, padx=10, pady=10, sticky="e")

        #listbox for seria port
        self.serial_port_listbox = Listbox(self, selectmode="single", exportselection=False)
        self.serial_port_listbox.grid(row=6, column=0, columnspan=2, padx=10, pady=5, sticky="nsew")
        self.serial_port_scrollbar = Scrollbar(self, orient="vertical", command=self.serial_port_listbox.yview)
        self.serial_port_scrollbar.grid(row=6, column=1, padx=(0, 10), pady=5, sticky="nes")
        self.serial_port_listbox.config(yscrollcommand=self.serial_port_scrollbar.set)

        #flash button
        self.flash_device_btn = Button(self, text="Flash Device")
        self.flash_device_btn.grid(row=7, column=0, padx=10, pady=10, sticky="w")

        #label for success, fail, inprogress, idle
        self.status_text = [("", "black"), ("Fail", "red"), ("Success", "green"), ("In Progress", "black")]
        i = 0
        self.status_label = Label(self, text=self.status_text[i][0], fg=self.status_text[i][1])
        self.status_label.grid(row=7, column=1, padx=10, pady=10, sticky="ew")

        self.info_label = Label(self, text="🛈 flashing the microcontroller might take a while. (approximately  1-2 mins)")
        self.info_label.grid(row=8, column=0, padx=10, pady=10, sticky="ew")

        self.wifi_profile_listbox.bind("<FocusIn>", lambda event: self.wifi_profile_listbox.selection_clear(0, tk.END))
        self.serial_port_listbox.bind("<FocusIn>", lambda event: self.serial_port_listbox.selection_clear(0, tk.END))


if __name__ == "__main__":
    root = tk.Tk()
    root.title("Add New Device")
    root.geometry("400x620")
    add_device_view = AddDeviceView(root)
    add_device_view.pack(fill="both", expand=True)
    root.mainloop()



from typing import Callable
from typing import List, TypeVar, Generic, Dict
from queue import Queue

from .Device import Device
from ..Models.SensorCapabilities import SensorCapabilities
from ..Models.BaseDataModel import BaseDataModel
from ..MayakoData import MayakoData
from ..Config import CCLIENT

#a generic type which is a subclass of BaseDataModel - a sensor data model class
T = TypeVar("T", bound=BaseDataModel)

class Sensor(Generic[T], Device):

    """
    this class represents a sensor on the remote microcontroller

    this class lets the user read and modify the capabilities of a sensor and subscribe OR/AND read sensor data. this class may not directly be instantiated! use "use_sensor()" method from the microcontroller class instead.

    Attributes:
        _data (Queue[T]): holds data until a user consumes it; uses the modelclass that is passed when this sensor was created
        _callback_functions (List[Callable[[T], None]]): the callback functions called when new data is available; function parameter is of type T
        _sensor_data_callback_identity (str): this is a unique key throughout the project which consists of the terms DATA, MC_IDENTITY and SENSOR_IDENTITY. this is used for the model class to specifically call this instance when new sensor data is available for it.
    """

    _data: Queue[T]
    _callback_functions: List[Callable[[List[T]], None]]
    _sensor_data_callback_identity: str

    def __init__(self, sensor_identity: str, mc_identity: str, model: MayakoData, sensor_class_type: T) -> None:
        super().__init__(identity=sensor_identity, mc_identity=mc_identity, model=model)
        self._sensor_class_type = sensor_class_type
        self._check_capabilities(sensor_identity=self._identity)
        self._sensor_data_callback_identity = f"DATA_{self._mc_identity}_{self._identity}"
        self._data = Queue()
        self._callback_functions = []
        self._subscribe_to_data()

    def get_data(self) -> List[T]:
        """
        retrieve data that is associated with this sensor if available

        this method provides the first option to access sensor data. this sensor subscribes on the model which executes the _data_callback if there is an update. the new data is updated to _data. when the user executes get_data(), it returns all data and clears the list, so that the users does not get the same data twice.

        Returns:
            List[T]: a list of data of type T
        """
        current_data = []

        while not self._data.empty():
            current_data.append(self._data.get())
        
        #first, i just made a copy of the list, cleared the ist and returned the data for the user. but i suspected race conditions in the time the data is transfered to the new list and before clearing the data from the _data list. if there was new data from the callback, we would accidently delete it, not knowing that it was there. but i am not sure if the event loop of python works like this. now we use a queue that is popped as long as it is not empty. safe not to lose data.

        return current_data
    
    def subscribe(self, callback: Callable[[T], None]) -> None:
        """
        subscribe to the sensor data that is updated when there is new data

        this methods registers a callback function from the user which is called when there is a sensor data update.

        Args:
            callback (Callable[[T], None]): a callback function which has the parameter of type T
        """
        if not callable(callback):
            raise TypeError(CERROR_MESSAGES.TYPE_MUST_BE_CALLABLE)

        try:
            self._callback_functions.append(callback)
        except:
            self._callback_functions = [callback]

    def _notify_callback_functions(self, data: T) -> None:
        """
        this function calls all callback functions when there is a data update in the model class

        this is the function which calls the callback function that is passed in subscribe()
        
        Args:
            data (T): a data point from type T
        """
        for callback in self._callback_functions:
            callback(data)

    def _data_callback(self, mayako_data: MayakoData) -> None:
        """
        this function is called when the model class has a new data point. As the get_sensor_data method from model returns a Dict, we also need to create an instance of the with this class associated model class. it updates the _data list and notifies all callback functions.

        Args:
            mayako_data (MayakoData): the model class with the data
        """
        data = mayako_data.get_sensor_data(self._mc_identity, self._identity)

        for datapoint in data:
            self._data.put(self._sensor_class_type.from_dict(datapoint))

        self._notify_callback_functions(data)

    def _subscribe_to_data(self) -> None:
        """this method subscribe the sensor to model data with typ associated in the event key"""
        self._model.add_sensor_to_data_list(self._mc_identity, self._identity)
        self._model.subscribe(self._sensor_data_callback_identity, self._data_callback)

    def define_capabilities(self, enable: bool = True, sample_rate: int = 10, data_on_state_change: bool = False) -> None:
        """
        this function lets the user change the capabilities of this sensor

        first, we check if there is a capabiltity that matches the identtiy of the microcontroller. then we create a new sensor capability object which

        Sources:
            K Yu (2016) Sampling interval and sampling rate/frequency - YouTube. Retrieved September 05, 2024, from https://www.youtube.com/watch?v=_iF-QNrZCU4&t=317s
            https://youtu.be/_iF-QNrZCU4?si=4WGSlzL2vugy0uwA

        Args:
            enable (bool): enables/disables the sensor in the record
            sample_rate (int): this is the times per second a data point is read from the microcontroller. check the video under Sources if the relation between sample rate and interval is not clear.
            data_on_state_change (bool): if True, the newly read data point on the microcontroller gets checked if it matches with the previous data point. if it does, the new data point is droped. no data point is sent from the microcontroller.
        """
        old_capas = self._check_capabilities(sensor_identity=self._identity)
        
        new_capas = SensorCapabilities(identity=self._identity, sample_rate=sample_rate, data_on_state_change=data_on_state_change, enable=enable)

        if not self._compare_sensor_capabilities(old_capas=old_capas, new_capas=new_capas):
            #we dont need to update the new capability because unlike with microcapabilities we
            self._logger.debug("updating sensor capability")
            self._model.update_sensor_capability(self._mc_identity, new_capas)
        else:
            self._logger.debug("no update on sensor capability")

    def _compare_sensor_capabilities(self, old_capas: SensorCapabilities, new_capas: SensorCapabilities) -> bool:
        """compares the sensor capabilities"""
        return (
            old_capas.enable == new_capas.enable and
            old_capas.sample_rate == new_capas.sample_rate and
            old_capas.data_on_state_change == new_capas.data_on_state_change
        )

    def _check_capabilities(self, sensor_identity: str) -> SensorCapabilities:
        """
        checks if there is a capability associated with the sensor identity
        
        Args:
            sensor_identity (str): identity of the sensor
        """
        capas = self._model.get_sensor_capability_by_identity(mc_identity=self._mc_identity, sensor_identity=sensor_identity)
        if not capas:
            raise Exception(f"the sensor with the provided identity {sensor_identity} does not exist in {CCLIENT.CONFIG_FILE_NAME}.")
        
        return capas

    def get_capabilities(self) -> SensorCapabilities:
        """returns the capabilities of the sensor"""
        return self._model.get_sensor_capability_by_identity(self._identity)


from typing import Dict, Callable, List

from .Utils.Logger import LoggerInstance, LoggerType


class Observable:

    _observers: Dict[str, List[Callable]]
    _logger: LoggerType

    def __init__(self):
        self._observers = {}
        self._logger = LoggerInstance.get()

    def subscribe(self, event: str, callback: Callable) -> None:
        if not callable(callback):
            raise TypeError(CERROR_MESSAGES.TYPE_MUST_BE_CALLABLE)

        try:
            self._observers[event].append(callback)
        except:
            self._observers[event] = [callback]

    def _notify_observers(self, event):
        if not event in self._observers.keys():
            return

        for callback in self._observers[event]:
            callback(self)


import os
from typing import Callable, List, Dict
from queue import Queue
from enum import Enum
from pathlib import Path
from pynput.keyboard import Listener
import time
import threading
import json
import asyncio

from .Network import NetworkManager
from .Network.NetworkInterface import NetworkProtocols
from .Utils.Logger import LoggerInstance, LoggerType, LoggingLevel
from .Config import CCLIENT, CGENERAL, CNETWORK, CCOMMAND
from .MayakoData import MayakoData
from .Devices.MicroController import MicroController
from .Devices.Sensor import Sensor
from .Devices.Actuator import Actuator
from .Command.CustomCommands import CustomCommands
from .Command.StandardCommands import StandardCommands
from .Network.Packet import Packet
from .Models.Button import Button
from .Service.BLEScanner import scan_ble_addresses
from .Service.SerialScanner import scan_serial_ports
from .Models.MicrocontrollerCapabilities import MicroControllerCapabilities
from .Network.BLE import BLE
from .Config import CPACKET
from .Network.NetworkManager import RecordStatus

class FileType(Enum):
    JSON="JSON"
    CSV="CSV"

class Client:

    _logger: LoggerType
    _port: int
    _debug: bool
    _path: str
    _config: dict
    _model: MayakoData
    _microcontrollers: Dict[str, NetworkManager]
    _sensors: List[Sensor]
    _actuators: List[Actuator]
    _commands: StandardCommands
    _custom_commands: CustomCommands
    _outgoing_network_queue: Queue[Packet]

    def __init__(self, port: int = CCLIENT.PORT, debug: bool = False, enable_ack: bool = False) -> None:
        self._port = port
        self._debug = debug
        self._debug = True
        self._enable_ack = enable_ack
        if CGENERAL.DEBUG or self._debug:
            LoggerInstance.init_logger(logging_level=LoggingLevel.DEBUG)
        else:
            LoggerInstance.init_logger()
        self._logger = LoggerInstance.get()
        self._model = MayakoData()
        self.record_status = RecordStatus.RECORD_INIT
        self.running = True
        self._run_interval = 0.01
        self._outgoing_network_queue = Queue()
        self._incoming_network_queue = Queue()
        self._microcontrollers = {}
        self._commands = StandardCommands(self._outgoing_network_queue)
        self._custom_commands = CustomCommands(self._outgoing_network_queue)
        self._start_callback = None
        self._stop_callback = None

        #now we must add bluetooth devices automatically when they are added but it is better when this happens in the thread
        self._thread = threading.Thread(target=self._run)
        self._thread.start()   

    def register_key_for_exit(self, key: str = 'q') -> None:
        self._keyboard_listener = Listener(on_press=self._keyboard_callback)
        self._keyboard_listener.start()
        self._keyboard_key = key
    
    def _keyboard_callback(self, key) -> None:
        if hasattr(key, 'char') and key.char == self._keyboard_key:
            self._logger.debug("stopping record...")
            self._keyboard_listener.stop()
            self.stop_record()

    def use_microcontroller(self, mc_identity: str) -> MicroController:
        mc_capa = self._model.get_microcontroller_capa_by_identity(identity=mc_identity)
        if not mc_capa:
            raise Exception(f"the microcontroller with the provided identity {mc_identity} does not exist in {CCLIENT.CONFIG_FILE_NAME}. add a new device with the according identity in mayako-gui before proceeding.")
       
        mc = MicroController(mc_identity=mc_identity, model=self._model, custom_commands=self._custom_commands)
        
        mc_capa = mc.get_capabilities()
        ble_address = self._model.get_ble_address_by_identity(mc_identity=mc_identity)
        nm = NetworkManager(mc_identity=mc_identity)
        self._microcontrollers[mc_identity] = nm
        self._microcontrollers[mc_identity].use_ble(ble_address=ble_address)
        self._connect_microcontroller(mc_identity=mc_identity)

        return mc
    
    def _scan_ble_address(self, mc_identity: str) -> None:
        self._logger.debug("scanning ble addresses")
        ble_address = asyncio.run(scan_ble_addresses(mc_identity, timeout=2))
        if not ble_address:
            raise Exception(f"no BLE device with identity {mc_identity} found")
        
        self._model.add_ble_address([ble_address])
        self._logger.debug(self._model.ble_addresses)

    def _connect_microcontroller(self, mc_identity: str) -> None:
        self._logger.debug("connecting...")
        if mc_identity in self._microcontrollers and not self._microcontrollers[mc_identity].check_connection():            
            connect_thread = threading.Thread(target=self._microcontrollers[mc_identity].connect())
            connect_thread.start()

    def _disconnect_microcontroller(self, mc_identity: str) -> None:
        if mc_identity in self._microcontrollers and self._microcontrollers[mc_identity].check_connection():
            self._logger.debug("disconnecting...")
            disconnect_thread = threading.Thread(target=self._microcontrollers[mc_identity].disconnect())
            disconnect_thread.start()

    def start_gui(self) -> None:
        from .GUI.Views.MainView import MainView
        from .GUI.Controllers.MainController import MainController

        main_view = MainView()
        main_controller = MainController(main_view, self._model, self)
        
        main_controller.start()

    def save_data(self, file_path: str, file_type: FileType=FileType.JSON) -> None:
        self._logger.debug("saving file")
        """
        writes all sensor data to a file

        Todo:
            implement data dump to csv

        Sources:
            https://stackoverflow.com/a/62662388
        
        Args:
            file_path (str): the location where the file should be written; if the location is invalid, the Downloads folder will be chosen as a fallback solution
            file_type (FileType): choose the file format of the data file; supports csv and json
        """
        if file_type == FileType.JSON:
            all_data = self._model.dump_all_sensor_data_to_json()
            file_name = "data.json"
        else:
            raise NotImplementedError

        if not os.path.exists(file_path):
            fallback_path = str(Path.home() / "Downloads")

            if not os.path.exists(fallback_path):
                raise Exception(f"{file_path} does not exist. fallback soluation {fallback_path} did also not work")

            else:
                file_path = os.path.join(fallback_path, file_name)
        else:
            file_path = os.path.join(file_path, file_name)
            
        with open(file_path, "w") as file:
            file.write(all_data)
            file.close()
            self._logger.info(f"data written to {file_path}")

    def _run(self) -> None:
        while self.running:
            
            for _, mc in self._microcontrollers.items():
                
                if mc._network_implementation == None:#TODO; we may not run on as long as ble is not setup
                    continue

                packet = mc.read_data()

                if packet and packet.get_method() == CPACKET.METHOD_DATA:
                    self._model.add_new_sensor_data(packet)

                if packet and packet.get_method() == CPACKET.METHOD_INFO:
                    data = json.loads(packet.get_payload())
                    data["identity"] = packet.get_node_identity()
                    if "name" in data and data["name"] == CCOMMAND.RECORD_READ:
                        self._logger.debug(data["name"])
                        mcCapa = MicroControllerCapabilities.from_dict(data)
                        self._model.update_microcontroller_capa(mcCapa)
            
            while not self._outgoing_network_queue.empty():
                output = self._outgoing_network_queue.get()
                self._logger.debug(output.get_payload())
                self._microcontrollers[output.get_node_identity()].send_data(output)

            time.sleep(self._run_interval)

    def start_record(self) -> None:
        self.running = True
        for identity, _ in self._microcontrollers.items():
            self._commands.record_start(mc_identity=identity)   

        if self._start_callback:
            self._start_callback()

    def stop_record(self) -> None:
        self._logger.debug("calling stop record")
        """this method stop the microcontrollers and calls the callback function register in on_stop"""
        
        self._disconnect_microcontrollers()       
        self._close_client_thread()

    def _disconnect_microcontrollers(self) -> None:
        for identity, _ in self._microcontrollers.items():
            self._commands.record_stop(mc_identity=identity)
            self._disconnect_microcontroller()

    def _close_client_thread(self) -> None:
        if self.running:
            self.running = False
            self._thread.join()

        if self._stop_callback:
            self._stop_callback()

    def on_start(self, callback: Callable[[], None]) -> None:
        """registers a callback function that is called when the record has started"""
        self._start_callback = callback

    def on_stop(self, callback: Callable[[], None]) -> None:
        """registers a callback function that is called when all microcontrollers have stopped sending data"""
        self._stop_callback = callback



from dataclasses import dataclass, asdict
from typing import Self, Dict

@dataclass
class WiFiProfile:
    """
    the is the model class for wifi profiles

    the wifi profile is setup in the GUI and can be appended as a build flag when flashing the arduino board using wireless_mode WIFI.

    Attributes:
        wifi_key (str): a short identity for a wifi profile; must be unique
        ssid (str): the address of the wifi router
        password (str): the password the accesses the wifi
        client_ip (str): the IP of the client application available by the router and thereby from the microcontroller
        client_port (int): the bound port on the clients machine to access the socket
    
    """

    wifi_key: str
    ssid: str
    password: str
    client_ip: str
    client_port: int

    def format_string(self) -> str:
        """
        this method formats a string from selected properties of the class to return a string that can be displayed in the GUI

        Returns:
            (str): a string describing the wifi profile
        """
        return f"WiFi Key: {self.wifi_key} -- SSID: {self.ssid} -- client IP: {self.client_ip} -- client Port: {self.client_port}"
    
    def to_dict(self) -> Dict:
        """
        returns the wifi profile as a Dict

        Returns:
            (Dict): wifi profile as dict
        """
        return asdict(self)
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """
        converts a dictionary to a class of the according type

        This is a classmethod which converts a dictionary that contains all class properties into an instance of the according type. the dictionary can contain properties that are not fitting in the class; theses properties will be ignored. This is helpful to filter out information that is not part of this class.

        Sources:
            https://stackoverflow.com/a/73264198
            https://stackoverflow.com/a/54769644

        Args:
            data (Dict): a dictionary that contains all properties of this class and more

        Returns:
            Self: return an instance of the class that contains the values from the dictionary

        """
        return cls(**data)


from tkinter import Frame, Label, Button, Entry, Text, Scrollbar, VERTICAL
import tkinter as tk

class DownloadArduinoView(Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.grid_columnconfigure(0, weight=0)
        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(9, weight=1) 

        #headline
        self.headline = Label(self, text="Download Arduino Project")
        self.headline.grid(row=0, column=0, columnspan=2, padx=10, pady=10, sticky="w")

        #return button
        self.return_btn = Button(self, text="Return")
        self.return_btn.grid(row=0, column=1, padx=10, pady=10, sticky="e")

        # GitHub Repository Entry
        self.repo_url_label = Label(self, text="Github Repository URL")
        self.repo_url_label.grid(row=1, column=0, padx=10,  sticky="w")
        self.repo_url_entry = Entry(self)
        self.repo_url_entry.grid(row=2, column=0, columnspan=2, padx=10, pady=10, sticky="we")
        self.repo_url_entry.insert(0, "https://github.com/vairasza/mayako-node")

        #GitHub Access Token which helps increase the rate limit drastically
        self.gh_token_label = Label(self, text="Github Access Token (optional)")
        self.gh_token_label.grid(row=3, column=0, padx=10,  sticky="w")
        self.gh_token_entry = Entry(self)
        self.gh_token_entry.grid(row=4, column=0, columnspan=2, padx=10, pady=10, sticky="we")

        # Destination Entry
        self.destination_label = Label(self, text="Target Folder for the Arduino Project")
        self.destination_label.grid(row=5, column=0, padx=10,  sticky="w")
        self.destination_entry = Entry(self)
        self.destination_entry.grid(row=6, column=0, columnspan=2, padx=10, pady=10, sticky="we")
        self.destination_entry.insert(0, "arduino")

        #flash button
        self.download_btn = Button(self, text="Download")
        self.download_btn.grid(row=7, column=0, padx=10, pady=5, sticky="w")

        #label for success, fail, inprogress, idle
        self.status_text = [("", "black"), ("Fail", "red"), ("Success", "green"), ("In Progress", "black")]
        self.status_label = Label(self, text=self.status_text[0][0], fg=self.status_text[0][1])
        self.status_label.grid(row=7, column=1, padx=10, pady=10, sticky="ew")

        self.download_hint = Label(self, text="🛈 After downloading the Arduino Project, check the repository's documentation if you intend to modify the code or want to extend it with additional sensors or actuators. Happy coding!", wraplength=650, justify="left")
        self.download_hint.grid(row=8, column=0, columnspan=2, padx=10, pady=10, sticky="w")

        #output for result of download
        self.result_text = Text(self)
        self.result_text.grid(row=9, column=0, columnspan=2, padx=(10, 0), pady=5, sticky="ew")
        self.result_scrollbar = Scrollbar(self, orient=VERTICAL, command=self.result_text.yview)
        self.result_scrollbar.grid(row=9, column=2, padx=(0, 10), pady=5, sticky="nesw")
        self.result_text.config(yscrollcommand=self.result_scrollbar.set)

if __name__ == "__main__":
    root = tk.Tk()
    root.title("Add New WiFi Profile")
    root.geometry("450x600")
    add_wifi_profile_view = DownloadArduinoView(root)
    add_wifi_profile_view.pack(fill="both", expand=True)
    root.mainloop()



from abc import ABC
from dataclasses import dataclass, asdict
from typing import Self, Dict, Optional

@dataclass
class BaseDataModel(ABC):
    """
    this is the base class for all model classes that are related to sensor data from the microcontrollers such as Accelerometer or Heartratesensor. Creating new model classes requires inheriting this class.

    Sources:
        https://www.arduino.cc/reference/en/language/functions/time/millis/

    Attributes:
        identity (str): the unique identity within a microcontroller that indicates the sensor
        timestamp (Optional[int]): a timestamp when the data point was created; pay attention that arduino boards use millis() function which returns the milliseconds passed from starting the arduino board, not unix time; is optional; default is None
        sequence (Optional[int]): the sequence of data points since the record started; is optional; default is None

    Source
        https://stackoverflow.com/a/2544761
    """
    identity: str
    timestamp: Optional[int]
    sequence: Optional[int]

    def to_dict(self, filter_none=True) -> dict:
        """
        Makes a dictionary from the model class.

        Args:
            filter_none (bool):
                all None values are excluded from the conversion

        Returns:
            Dict: a dictionary representation of this instance
        """
        data = asdict(self)

        if filter_none:
            return {key: value for key, value in data.items() if value is not None}
        
        return data
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """
        Creates a model instance from dictionary.

        this method creates an instance of the according model class using information from a dictionary. as the values timestamp and sequence may not be included, a default value is set to None.

        Sources:
            https://stackoverflow.com/a/3483652

        Returns:
            (Self): returns an instance of the model class
        """
        data.setdefault("timestamp", None)
        data.setdefault("sequence", None)
        
        return cls(**data)
    
    def get_class_name(self) -> str:
        """
        returns the name of the model class
        """
        return self.__class__.__name__


import tkinter as tk

class Root(tk.Tk):
    def __init__(self):
        super().__init__()

        start_width = 750
        min_width = 400
        start_height = 750
        min_height = 400

        self.geometry(f"{start_width}x{start_height}")
        self.minsize(width=min_width, height=min_height)
        self.title("mayako GUI")
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)


from dataclasses import dataclass, asdict
from typing import Dict, Self
@dataclass
class ActuatorCapabilities:
    """
    a model class which contains capabilities of actuators.

    the capabilities only contain parameters which influence the record behaviour plus identity to indentify the actuator instance.

    Attributes:
        enable (bool): if a actuator is enabled for the record; default is True because when the actuator is created by the user, it is obvious that he wants to enable it. the actuators on the device are set to disabled per default only being enabled when an according capability is passed.
        identity (str): the identity of the actuator

    """
    identity: str
    enable: bool = True
    commands: str = ""

    def to_dict(self) -> Dict:
        """makes a dictionary from the instance"""
        return asdict(self)
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """
        makes an instance of the according class using a dictionary.

        not existing values are set to default values. identity must exist as it is an integral part.

        Args:
            data (Dict): dictionary with instance properties

        Returns:
            Self: an instance of the class with the assigned values
        """
        default_data = {
            'enable': True,
            #identity must be provided
        }
        
        merged_data = {**default_data, **data}
        return cls(**merged_data)


import tkinter as tk
from ..Views.MainView import MainView
from ..Views.EditWiFiProfileView import EditWiFiProfileView
from ...MayakoData import MayakoData
from ...Models.WiFiProfile import WiFiProfile
from ...Client import Client

class EditWiFiProfileController:

    main_view: MainView
    edit_wifi_frame: EditWiFiProfileView

    def __init__(self, main_view: MainView, model: MayakoData, client: Client) -> None:
        self.main_view = main_view
        self.model = model
        self.client = client
        self.edit_wifi_frame = self.main_view.frame_classes["edit_wifi"]
        self._bind()
        self._model_subscribe()

    def _bind(self) -> None:
        self.edit_wifi_frame.return_btn.config(command=self._return_to_main)
        self.edit_wifi_frame.edit_wifi_profile_btn.config(command=self._update_wifi_profile)

    def _return_to_main(self, *_) -> None:
        self.main_view.switch("home")

    def _model_subscribe(self) -> None:
        self.model.subscribe("wifi_profile_selected", self._load_wifi_profile)

    def _load_wifi_profile(self, mayako_data: MayakoData) -> None:
        wifi_profile = mayako_data.selected_wifi_profile
        
        self.edit_wifi_frame.wifi_key_entry.config(text=wifi_profile.wifi_key)

        self.edit_wifi_frame.ssid_entry.delete(0, tk.END)
        self.edit_wifi_frame.ssid_entry.insert(0, wifi_profile.ssid)

        self.edit_wifi_frame.password_entry.delete(0, tk.END)
        self.edit_wifi_frame.password_entry.insert(0, wifi_profile.password)

        self.edit_wifi_frame.client_ip_entry.delete(0, tk.END)
        self.edit_wifi_frame.client_ip_entry.insert(0, wifi_profile.client_ip)

        self.edit_wifi_frame.client_port_entry.delete(0, tk.END)
        self.edit_wifi_frame.client_port_entry.insert(0, wifi_profile.client_port)

    def _update_wifi_profile(self, *_) -> None:
        existing_wifi_key = self.model.selected_wifi_profile.wifi_key
        
        ssid = self.edit_wifi_frame.ssid_entry.get()
        password = self.edit_wifi_frame.password_entry.get()
        client_ip = self.edit_wifi_frame.client_ip_entry.get()
        client_port = int(self.edit_wifi_frame.client_port_entry.get())

        updated_wifi_profile = WiFiProfile(wifi_key=existing_wifi_key, ssid=ssid, password=password, client_ip=client_ip, client_port=client_port)

        self.model.update_wifi_profile(updated_wifi_profile)

        self.main_view.switch("home")



import socket

def get_user_ip_address() -> int:
    """
    this function creates a socket, connects to a random ip address, reads the socket information that also contains the machines used ip address.

    Sources:
        https://stackoverflow.com/a/30990617

    Returns:
        int: the ip address of the machine
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip = s.getsockname()[0]
    s.close()

    return ip


from bleak import BleakScanner, BleakClient
from typing import Optional

from ..Models.BLEInfo import BLEInfo

async def scan_ble_addresses(device_name: str, timeout: int = 5) -> Optional[BLEInfo]:
    """
    This function scans for devices that are advertising their BLE presence. The function is an async function using asyncio. Therefore the function asyncio.run(scan_ble_addresses(["test-device"])) must be used to correctly invoke this function. It first checks if the device_names are of type list, even if it is only one item to void problems with the check if the device name is in the list. BleakScanner scans for all devices with BLE returning device and advertisement informations without in timeout window of 5 seconds per default. this can be adjusted. if the device matches one of the device_names - or partly - it is appended the list of devices that we are looking for. the BLEinfor class is a dataclass which contains the information about the mac address, the name of the device (usually a human readable string) and the service uuids - this is also a list because BLE servers can have multiple services.

    Source:
        https://github.com/hbldh/bleak
        https://github.com/hbldh/bleak/blob/develop/examples/discover.py
        https://bleak.readthedocs.io/en/latest/api/index.html#bleak.backends.characteristic.BleakGATTCharacteristic

    Args:
        device_names (List[str]):
            list of ble device names. the name must not be complete (it checks if any part of the device name matches one of the lists items) although this can lead to unwated ble devices in the list.
        timeout (int):
            the BLE scanner stops after x seconds (5 seconds per default) and returns the list of ble devices found

    Returns:
        list of ble devices which match the device names list
    
    Raises:
        TypeError: if the argument device_names is not of type list
    """    
    ble_address: Optional[BLEInfo] = None

    devices = await BleakScanner.discover(return_adv=True, timeout=timeout)

    for device, _ in devices.values():
        if device_name in device.name:
            async with BleakClient(device.address) as client:
                services = await client.get_services()

                ble_info = {}
                for service in services:
                    for char in service.characteristics:
                        if "write" in char.properties and "notify" in char.properties:
                            ble_info["service_uuid"] = service.uuid,
                            ble_info["characteristic_uuid"] = char.uuid

                if ble_info:
                    ble_address = BLEInfo(device.address, device.name, ble_info["service_uuid"], ble_info["characteristic_uuid"])
                    
                    break
            
    return ble_address



from .NetworkInterface import NetworkInterface

class IntegrityMiddleware:

    counter: list[any]
    protocol: NetworkInterface

    def __init__(self, protocol: NetworkInterface) -> None:
        self.protocol = protocol
        

    def write(self) -> None:
        self.protocol.write()

    def read(self) -> None:
        self.protocol.read()


from queue import Queue
from typing import Dict
import json

from ..Config import CCOMMAND, CPACKET
from ..Models.MicrocontrollerCapabilities import MicroControllerCapabilities
from ..Models.WiFiProfile import WiFiProfile
from ..Network.Packet import Packet

class StandardCommands:
    """
    this is the class for the standard commands which are defined in protocol_interface.md

    it is not meant to be directly used by the user. each function creates a Packet instance which holds the mc_identity (the destination of the packet), the command method, and the payload. the packet instance is stored in the queue of the client which is frequently flushed to the network.
    """

    def __init__(self, client_queue: Queue[Packet]) -> None:
        self._client_queue = client_queue

    def _make_packet(self, mc_identity:str, doc: Dict) -> None:
        """
        this is a helper function which creates a packet from the dictionary.

        Args:
            mc_identity (str): identity of a microcontroller
            doc (Dict): the payload of the packet in dictionary format

        :
            Packet:  a ready-to-send packet
        """
        jsonDoc = json.dumps(doc)

        packet = Packet()
        packet.set_method(CPACKET.METHOD_COMMAND)
        packet.set_node_identity(mc_identity)
        packet.set_payload(jsonDoc)
        
        self._client_queue.put(packet)

    def record_create(self, mc_identity: str, capabilities: MicroControllerCapabilities) -> None:
        """
        this method creates a packet from the provided model class and the identity of the microcontroller

        mc_identity is used so that we know where to send the packet to (mapping mc <-> network address). use packet.get_node_identity() to get the identity. adding cmd_name to the dictionary created with to_dict() and serialise as json string. the packet is created with the method which is fix COMMAND, the identity (microcontroller) and the payload which is the json string. this packet can directly be sent. we want the payload to be flat because this makes it easier on the mayako-node and we only have to add cmd_name (only keyword in the payload). also most of the time we dont have a payload. adding a parameters property would only bloat the payload.

        Args:
            mc_identity (str): identity of the microcontroller that is the target of the packet
            capabilities(MicroControllerCapabilities): capabilities of a certain microcotrnoller

        :
            Packet: a packet class that contains all information for a network implementation to send it to the micrcontroller
        """
        #mc/node identiy so that we know where this packet should be aimed at
        doc = capabilities.to_dict()
        doc["cmd_name"] = CCOMMAND.RECORD_CREATE
        
        self._make_packet(mc_identity, doc)
    
    def record_start(self, mc_identity: str) -> None:
        """
        this command start the record on the mayako-node if it is not already running

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.RECORD_START

        self._make_packet(mc_identity, doc)

    def record_stop(self, mc_identity: str) -> None:
        """
        this command stops the record on the mayako-node if it is running

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.RECORD_STOP

        self._make_packet(mc_identity, doc)

    def record_read(self, mc_identity: str) -> None:
        """
        this command reads the capabilities of the mayako-node

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.RECORD_READ

        self._make_packet(mc_identity, doc)

    def restart(self, mc_identity: str) -> None:
        """
        this command restarts the mayako-node

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.RESTART

        self._make_packet(mc_identity, doc)

    def battery_read(self, mc_identity: str) -> None:
        """
        this command reads the battery of the mayako-node

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.BATTERY_READ

        self._make_packet(mc_identity, doc)

    def identify(self, mc_identity: str, identity: str) -> None:
        """
        this command start the record on the mayako-node if it is not already running

        Args:
            mc_identity (str): identity of the microcontroller
            identity (str): this is the identity of the microcontroller/sensor/actuator which should be identified.

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.IDENTIFY
        doc["identity"] = identity

        self._make_packet(mc_identity, doc)

    def connection_read(self, mc_identity: str) -> None:
        """
        this command reads the connection status and the protocol used

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.CONNECTION_READ

        self._make_packet(mc_identity, doc)

    def wifi_profile_create(self, mc_identity: str, wifi_profile: WiFiProfile) -> None:
        """
        this command creates a new wifi profile on the mayako-node

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = wifi_profile.to_dict()
        doc["cmd_name"] = CCOMMAND.WIFI_PROFILE_CREATE

        self._make_packet(mc_identity, doc)

    def wifi_profile_read(self, mc_identity: str, wifi_key: str) -> None:
        """
        this command read a wifi profile by wifi_key

        Args:
            mc_identity (str): identity of the microcontroller
            wifi_key (str): identity of a wifi profile

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.WIFI_PROFILE_READ
        doc["wifi_key"] = wifi_key

        self._make_packet(mc_identity, doc)

    def wifi_profile_active_read(self, mc_identity: str) -> None:
        """
        this command reads the wifi profile that is marked as the active wifi profile. the active wifi profile is the profile that is loaded on start.

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.WIFI_PROFILE_ACTIVE_READ

        self._make_packet(mc_identity, doc)

    def wifi_profile_all_read(self, mc_identity: str) -> None:
        """
        this command reads all wifi profiles on the mayako-node

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.WIFI_PROFILE_ALL_READ

        self._make_packet(mc_identity, doc)

    def wifi_profile_active_select(self, mc_identity: str, wifi_key: str) -> None:
        """
        this command changes the wifi profile to a wifi profile that is stored on the mayako-node using the according wifi-key

        Args:
            mc_identity (str): identity of the microcontroller
            wifi_key (str): the wifi_key of the wifi profile on the mayako-node that we want to select as the active wifi profile

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.WIFI_PROFILE_ACTIVE_SELECT
        doc["wifi_key"] = wifi_key

        self._make_packet(mc_identity, doc)

    def wifi_profile_delete(self, mc_identity: str) -> None:
        """
        this command reads all wifi profiles on the mayako-node

        Args:
            mc_identity (str): identity of the microcontroller

        :
            Packet: command wrapped in packet class
        """
        doc = {}
        doc["cmd_name"] = CCOMMAND.WIFI_PROFILE_DELETE

        self._make_packet(mc_identity, doc)





from ..Utils.Identity import Identity
from ..MayakoData import MayakoData
from ..Utils.Logger import LoggerInstance, LoggerType

class Device:

    _identity: str
    _mc_identity: str
    _model: MayakoData
    _logger: LoggerType

    def __init__(self, identity: str, mc_identity: str, model: MayakoData) -> None:
        Identity.register(identity)
        self._identity = identity
        self._mc_identity = mc_identity
        self._model = model
        self._logger = LoggerInstance.get()

    def get_identity(self) -> str:
        """returns the identity of the actuator or sensor"""
        return self._identity


import time
import socket
import threading
from queue import Queue
from typing import Optional

from .NetworkInterface import NetworkInterface
from ..Config import CNETWORK, CPACKET
from .Packet import Packet
from ..Utils.Logger import LoggerInstance, LoggerType

class WiFiUDP(NetworkInterface):

    _logger: LoggerType
    _server_ip: str
    _server_port: int
    _buffer_size: int
    _socket: socket.socket
    _read_queue: Queue[Packet]
    _write_queue: Queue[Packet]
    _stop_event: threading.Event
    _read_thread: threading.Thread
    _write_thread: threading.Thread

    def __init__(self, server_ip: str, server_port: int, buffer_size: int = CNETWORK.BUFFER_SIZE) -> None:
        self._logger = LoggerInstance.get()
        self._server_ip = server_ip
        self._server_port = server_port
        self._buffer_size = buffer_size
        self._socket = None
        self._read_queue = Queue()
        self._write_queue = Queue()
        self._stop_event = threading.Event()

    def connect(self) -> None:
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            self._socket.bind((self._server_ip, self._server_port))
        except socket.error as e:
            if e.errno == CNETWORK.PORT_ALREADY_USED_CODE:
                self._logger.error("port already in use")
            else:
                self._logger.error(f"error connecting to socket: {e}")
            
            return

        self._stop_event.clear()

        self._read_thread = threading.Thread(target=self._read_data)
        self._write_thread = threading.Thread(target=self._write_data)

        self._read_thread.start()
        self._write_thread.start()

    def _read_data(self):
        while not self._stop_event.is_set():
            try:
                if self._sock:
                    data, _ = self._sock.recvfrom(self._buffer_size)
                    if len(data) > CPACKET.HEADER_SIZE:
                        packet = Packet()
                        packet.deserialize_header(data[:CPACKET.HEADER_SIZE])
                        packet.deserialize_payload(data[CPACKET.HEADER_SIZE:])
                        self._read_queue.put(packet)
            except socket.error as e:
                self._logger.error(f"UDP error during read: {e}")
                break

            time.sleep(CNETWORK.BUSY_WAITING_TIMEOUT)

    def _write_data(self):
        while not self._stop_event.is_set():
            try:
                if not self._write_queue.empty():
                    packet = self._write_queue.get()
                    serialized_data = packet.serialize()

                    if self._sock:
                        self._sock.sendto(serialized_data, (self._udp_ip, self._udp_port))
            except socket.error as e:
                self._logger.error(f"UDP error during write: {e}")
                break

            time.sleep(CNETWORK.BUSY_WAITING_TIMEOUT)

    def disconnect(self) -> None:
        self._stop_event.set()

        if self._read_thread and self._read_thread.is_alive():
            self._read_thread.join()

        if self._write_thread and self._write_thread.is_alive():
            self._write_thread.join()

        if self._socket:
            self._socket.close()
            self._socket = None

    def write(self, data: bytes) -> None:
        self._write_queue.put(data)

    def read(self) -> bytes:
        if self._read_queue.qsize() > 0:
            return self._read_queue.get()

        return None


from .Accelerometer import Accelerometer
from .AHRS import AHRS
from .Button import Button
from .Gyroscope import Gyroscope
from .Heartrate import Heartrate
from .Temperature import Temperature

__all__ = ['Accelerometer', 'AHRS', 'Button', 'Gyroscope', 'Heartrate', 'Temperature', 'BaseDataModel']


from ..Config import CERROR_MESSAGES

class IdentitiyKeyUsedError(Exception):
    '''
    Expcetion raised when an identity key is used more than once.
    '''
    def __init__(self, identity_key: str) -> None:
        super().__init__(f'The identity key {identity_key} has already been used in another instance.')

class IdentitiyPatternError(Exception):
    '''
    Expcetion raised when an identity key is used more than once.
    '''
    def __init__(self, identity_key: str) -> None:
        super().__init__(f'The identity key {identity_key} must be between 2 and 8 characters long and may only contain underscores (_) as special characters.')


import json
from typing import List, Dict, Optional
import os

from .Observable import Observable
from .Models.WiFiProfile import WiFiProfile
from .Service.SerialScanner import SerialInfo
from .Service.BLEScanner import BLEInfo
from .Models.User import User
from .Models.MicrocontrollerCapabilities import MicroControllerCapabilities
from .Models.SensorCapabilities import SensorCapabilities
from .Models.ActuatorCapabilities import ActuatorCapabilities
from .Config import CCLIENT
from .Utils.Logger import LoggerType, LoggerInstance
from .Service.MacAdress import get_macaddress
from .Network.Packet import Packet
from .Models.BaseDataModel import BaseDataModel

class MayakoData(Observable):

    """
    this is the central model class in mayako-core observable to GUI and Client

    Attributes:
        logger (LoggerType): instance of logger class
        this_user (User): the current user identified by the machine MAC address
        all_users (List[User]): all users that have worked on different machines in this project
        wifi_profile (List[WiFiProfiles]): list of wifi profiles created in mayako GUI
        serial_ports (List[SerialInfo]): list of serial access info that is retrieved with SerialScanner.py
        ble_addresses (List[BLEInfo]): list of BLE addresses retrieved with BLEScanner.py
        mc_capabilities (List[MicroControllerCapabilities]): all microcontrollers capabilities
        data: Dict[str, Dict[str, Dict[str, List[Dict]]]]: sensor data that is sorted by mc_identity and sensor_identity
        selected_wifi_profile (WiFiProfie): the wifi_profile selected in GUI which is loaded into the EditWiFiProfileView
        selected_mc_capability (MicroControllerCapabilities): the MicroControllerCapabilities selected in GUI which is loaded into DeviceDetailsView

    Example:
        data: the data is cast to the associated model class when it is transfered to the sensors themselves
        
        .. code-block: python
            {
                "MC01": {
                    "TEMP01": {
                        "data": [{"identity": "TEMP_01","temperature": 24.22}],
                        "last_accessed_index": 0
                    }
                    "ACC01": []
                }
            }
    """

    logger: LoggerType
    this_user: User
    all_users: List[User]
    wifi_profiles: List[WiFiProfile]
    serial_ports: List[SerialInfo]
    ble_addresses: List[BLEInfo]
    mc_capabilities: List[MicroControllerCapabilities]#sensor and actuator capabilities are in mc_capabilities
    data: Dict[str, Dict[str, Dict[str, List[Dict]]]]
    selected_wifi_profile: WiFiProfile
    selected_mc_capability: MicroControllerCapabilities
    

    def __init__(self) -> None:
        super().__init__()
        self.logger = LoggerInstance.get()
        self.this_user = None
        self.all_users = []
        self.wifi_profiles = []
        self.serial_ports = []
        self.ble_addresses = []
        self.mc_capabilities = []
        self.data = {}
        self.selected_wifi_profile = None
        self.selected_mc_capability = None

        self._check_if_file_exists()
        self._load_from_file()
        self._check_if_user_exists()
        self._save_to_file()


        self._self_bind()

    def _check_if_file_exists(self) -> None:
        curr_workdir = os.getcwd()
        config_file_path = os.path.join(curr_workdir, CCLIENT.CONFIG_FILE_NAME)
        
        if not os.path.exists(config_file_path):
            with open(config_file_path, "w") as file:
                file.write("{}")
        else:
            self.logger.debug(f"{CCLIENT.CONFIG_FILE_NAME} already exists")

    def _check_if_user_exists(self) -> None:
        if not self.this_user:
            pass

    def _load_from_file(self) -> None:
        with open(CCLIENT.CONFIG_FILE_NAME, "r") as file:
            data = json.load(file)
            
            if "wifi_profiles" in data:
                # Load wifi profiles
                wifi_profiles = [WiFiProfile.from_dict(wifi_profile) for wifi_profile in data["wifi_profiles"]]
                self.add_wifi_profile(wifi_profiles)
            
            if "ble_addresses" in data:
                # Load BLE addresses
                ble_addresses = [BLEInfo.from_dict(item) for item in data["ble_addresses"]]
                self.add_ble_address(ble_addresses)
            
            if "microcontroller_capas" in data:
                # Load microcontrollers
                microcontroller_capas = [MicroControllerCapabilities.from_dict(item) for item in data["microcontroller_capas"]]
                self.add_microcontroller_capa(microcontroller_capas)
            
            if "users" in data:
                # Load users
                self.all_users = [User.from_dict(u) for u in data["users"]]
            
            self.this_user = self._check_user(self.all_users)

    def _save_to_file(self, *_) -> None:
        d = {
            "users": [],
            "wifi_profiles": [],
            "ble_addresses": [],
            "microcontroller_capas": []
        }

        for user in self.all_users:
            d["users"].append(user.to_dict())
        
        for wifi_profile in self.wifi_profiles:
            d["wifi_profiles"].append(wifi_profile.to_dict())

        for ble_address in self.ble_addresses:
            d["ble_addresses"].append(ble_address.to_dict())

        for microcontroller_capa in self.mc_capabilities:
            d["microcontroller_capas"].append(microcontroller_capa.to_dict())

        with open(CCLIENT.CONFIG_FILE_NAME, "w") as file:
            json.dump(d, file, indent=4)

    def _self_bind(self) -> None:
        self.subscribe("wifi_profiles_update", self._save_to_file)
        self.subscribe("microcontroller_capa_update", self._save_to_file)
        self.subscribe("serial_ports_update", self._save_to_file)
        self.subscribe("user_update", self._save_to_file)
        self.subscribe("ble_addresses_update", self._save_to_file)

    def _check_user(self, users: List[User]) -> User:
        user_address = get_macaddress()
        user = None
    
        for u in users:
            if u.user_mac_address == user_address:
                user = u
                return user
        
        new_user = User(user_mac_address=user_address, arduino_folder=CCLIENT.DEFAULT_ARDUINO_FOLDER)

        if new_user not in self.all_users:
            self.all_users.append(new_user)
            
        return new_user

    def add_wifi_profile(self, wifi_profiles: List[WiFiProfile]) -> None:
        for wifi_profile in wifi_profiles:
            if wifi_profile in self.wifi_profiles:
                continue
            
            self.wifi_profiles.append(wifi_profile)

        self._notify_observers("wifi_profiles_update")

    def remove_wifi_profile(self, wifi_profile: WiFiProfile) -> None:
        self.wifi_profiles.remove(wifi_profile)

        self._notify_observers("wifi_profiles_update")

    def get_wifi_profiles(self) -> List[WiFiProfile]:
        return self.wifi_profiles
    
    def select_wifi_profile(self, wifi_profile: WiFiProfile) -> None:
        if wifi_profile not in self.wifi_profiles:
            self.select_wifi_profile = self.wifi_profiles[0]
            self.logger.error("unknown wifi_profile selected")
        
        self.selected_wifi_profile = wifi_profile
        self._notify_observers("wifi_profile_selected")

    def get_wifi_profile_by_wifi_key(self, wifi_key: str) -> WiFiProfile:
        for wifi_profile in self.wifi_profiles:
            if wifi_profile.wifi_key == wifi_key:
                return wifi_profile
            
        return None

    def update_wifi_profile(self, updated_wifi_profile: WiFiProfile) -> None:
        for index, wifi_profile in enumerate(self.wifi_profiles):
            if updated_wifi_profile.wifi_key == wifi_profile.wifi_key:                
                self.wifi_profiles[index] = updated_wifi_profile
                break

        self._notify_observers("wifi_profiles_update")

    def add_microcontroller_capa(self, mcCapas: List[MicroControllerCapabilities]) -> None:
        for mcCapa in mcCapas:
            if mcCapa in self.mc_capabilities:
                continue
        
            self.mc_capabilities.append(mcCapa)

        self._notify_observers("microcontroller_capa_update")
    
    def remove_microcontroller_capa(self, mcCapa: MicroControllerCapabilities) -> None:
        self.mc_capabilities.remove(mcCapa)

        self._notify_observers("microcontroller_capa_update")

    def get_microcontroller_capas(self) -> List[MicroControllerCapabilities]:
        return self.mc_capabilities
    
    def select_microcontroller_capa(self, mcCapa: MicroControllerCapabilities) -> None:
        if mcCapa not in self.mc_capabilities:
            self.select_microcontroller = self.mc_capabilities[0]
            self.logger.error("unknown microcontroller capability selected")
        
        self.selected_mc_capability = mcCapa
        self._notify_observers("microcontroller_capa_selected")

    def get_microcontroller_capa_by_identity(self, identity: str) -> Optional[MicroControllerCapabilities]:
        for mcCapa in self.mc_capabilities:
            if mcCapa.identity == identity:
                return mcCapa
            
        return None
    
    def update_microcontroller_capa(self, mcCapa: MicroControllerCapabilities) -> None:
        for key, mc in enumerate(self.mc_capabilities):
            if mc.identity == mcCapa.identity:
                self.mc_capabilities[key] = mcCapa
                
                self._notify_observers("microcontroller_capa_update")

                break
        if self.selected_mc_capability and self.selected_mc_capability.identity == mcCapa.identity:
            self.selected_mc_capability = mcCapa
            self._notify_observers("microcontroller_capa_selected")

    def add_serial_port(self, serial_ports: List[SerialInfo]) -> None:
        self.serial_ports.clear()#to avoid doubles
        for serial_port in serial_ports:
            if serial_port in self.serial_ports:
                continue
        
            self.serial_ports.append(serial_port)
        
        self._notify_observers("serial_ports_update")

    def remove_serial_port(self, serial_port: SerialInfo) -> None:
        self.serial_ports.remove(serial_port)

        self._notify_observers("serial_ports_update")

    def get_serial_ports(self) -> List[SerialInfo]:
        return self.serial_ports
    
    def add_user(self, user: User) -> None:
        self.this_user.arduino_folder = user.arduino_folder

        self._notify_observers("user_update")

    def get_all_users(self) -> List[User]:
        return self.all_users
    
    def get_this_user(self) -> Optional[User]:
        return self.this_user
    
    def add_ble_address(self, ble_addresses: List[BLEInfo]) -> None:
        for ble_address in ble_addresses:
            if ble_address in self.ble_addresses:
                continue

            self.ble_addresses.append(ble_address)
        #TODO: check if we have doubles again
        self._notify_observers("ble_addresses_update")
    
    def remove_ble_address(self, ble_address: BLEInfo) -> None:
        self.ble_addresses.remove(ble_address)

        self._notify_observers("ble_addresses_update")

    def get_ble_addresses(self) -> List[BLEInfo]:
        return self.ble_addresses
    
    def get_ble_address_by_identity(self, mc_identity: str) -> Optional[BLEInfo]:
        for ble_address in self.ble_addresses:
            if ble_address.name == mc_identity:
                return ble_address
        
        return None
    
    def add_sensor_capabilities(self, mc_identity: str, sensor_capa: SensorCapabilities) -> None:
        raise NotImplementedError

    def remove_sensor_capability(self, mc_identity: str, sensor_capa: SensorCapabilities) -> None:
        raise NotImplementedError

    def get_sensor_capabilities(self, mc_identity: str) -> List[SensorCapabilities]:
        for mc_capa in self.mc_capabilities:
            if mc_capa.identity == mc_identity:
                return mc_capa.sensors
            
        return None
    
    def get_sensor_capability_by_identity(self, mc_identity: str, sensor_identity: str) -> Optional[SensorCapabilities]:
        for mc_capa in self.mc_capabilities:
            if mc_capa.identity == mc_identity:
                for sensor in mc_capa.sensors:
                    if sensor.identity == sensor_identity:
                        return sensor

        return None

    def update_sensor_capability(self, mc_identity: str, sensor_capa: SensorCapabilities) -> None:
        """
        this method replaces an old sensor capability by a new capability. be aware to include old capabilities that you are not updating so that we do not lose information.

        Args:
            mc_identity (str): the identity of the associated microcontroller
            sensor_capa (SensorCapabilities): the capabilities that want to be updated
        """
        for mc_capa in self.mc_capabilities:
            if mc_capa.identity == mc_identity:
                for key, sensor in enumerate(mc_capa.sensors):
                    if sensor.identity == sensor_capa.identity:
                        sensor[key] = sensor_capa
                        
                        self._notify_observers("microcontroller_capa_update")

                        break

    def add_actuator_capabilities(self, mc_identity: str, actuator_capa: ActuatorCapabilities) -> None:
        raise NotImplementedError

    def remove_actuator_capability(self, mc_identity: str, actuator_capa: ActuatorCapabilities) -> None:
        raise NotImplementedError
    
    def get_actuator_capabilities(self, mc_identity: str) -> List[ActuatorCapabilities]:
        for mc_capa in self.mc_capabilities:
            if mc_capa.identity == mc_identity:
                return mc_capa.actuators
            
        return None
    
    def get_actuator_capability_by_identity(self, mc_identity: str, actuator_identity: str) -> Optional[ActuatorCapabilities]:
        for mc_capa in self.mc_capabilities:
            if mc_capa.identity == mc_identity:
                for actuator in mc_capa.actuators:
                    if actuator.identity == actuator_identity:
                        return actuator

        return None
    
    def update_actuator_capability(self, mc_identity: str, actuator_capa: ActuatorCapabilities) -> None:
        """
        this method replaces an old actuator capability by a new capability. be aware to include old capabilities that you are not updating so that we do not lose information.

        Args:
            mc_identity (str): the identity of the associated microcontroller
            actuator_capa (ActuatorCapabilities): the capabilities that want to be updated
        """
        for mc_capa in self.mc_capabilities:
            if mc_capa.identity == mc_identity:
                for key, actuator in enumerate(mc_capa.actuators):
                    if actuator.identity == actuator_capa.identity:
                        actuator[key] = actuator_capa
                        
                        self._notify_observers("microcontroller_capa_update")

                        break

    def add_sensor_to_data_list(self, mc_identity: str, sensor_identity: str) -> None:
        """
        extend the _data object for additional microcontrollers and their associated sensors

        Args:
            mc_identity (str): the identity of a microcontroller
            sensor_identity (str): the identity of a sensor
        """
        if mc_identity not in self.data:
            self.data[mc_identity] = {}

        if sensor_identity not in self.data[mc_identity]:
            self.data[mc_identity][sensor_identity] = {}

        if "data" not in self.data[mc_identity][sensor_identity]:
            self.data[mc_identity][sensor_identity]["data"] = []

        if "last_accessed_index" not in self.data[mc_identity][sensor_identity]:
            self.data[mc_identity][sensor_identity]["last_accessed_index"] = -1

    def add_new_sensor_data(self, packet: Packet) -> None:
        """
        add new sensor data from the client to the model class

        this method retrieves microcontroller and sensor identity from header and payload, along with the payload itself. we convert the json to a dictionary and store it in self.data. we do NOT convert it to the associated modelclass because we do not know the model class while the sensor class does. and also, we can directly dump the dict to a file, whereas a datastructure with various model classes would be additional and unnessecary compute time.

        special attention to event_name:
        this is a unique key throughout the project which consists of the terms DATA, MC_IDENTITY and SENSOR_IDENTITY. this is used for this class to specifically call the sensor instance when new sensor data is available for it.

        Args:
            packet (Packet): packet that was sent from the network
        """
        mc_identity = packet.get_node_identity()
        json_data = packet.get_payload()
        data = json.loads(json_data)
        sensor_identity = data["identity"]
        event_name = f"DATA_{mc_identity}_{sensor_identity}"

        if mc_identity not in self.data:
            return
        
        if sensor_identity not in self.data[mc_identity]:
            return

        self.data[mc_identity][sensor_identity]["data"].append(data)
        
        self._notify_observers(event=event_name)

    def get_sensor_data(self, mc_identity: str, sensor_identity: str) -> List[Dict]:
        """
        this method is used to retrieve the latest sensor data, precisely cut to the sensor of a microcontroller
        """
        if mc_identity not in self.data or sensor_identity not in self.data[mc_identity]:
            # if the microcontroller or sensor identity does not exist, return an empty dict
            return []

        if "last_accessed_index" not in self.data[mc_identity][sensor_identity]:
            self.data[mc_identity][sensor_identity]["last_accessed_index"] = -1

        sensor_data = self.data[mc_identity][sensor_identity]
        last_accessed_index = sensor_data["last_accessed_index"]
        new_data = sensor_data["data"][last_accessed_index + 1:]

        if new_data:
            sensor_data["last_accessed_index"] = len(sensor_data["data"]) - 1

        return new_data
    
    def dump_all_sensor_data_to_json(self) -> None:
        all_data = {}

        for mc_identity, sensors in self.data.items():
            all_data[mc_identity] = {}
            for sensor_identity, sensor_info in sensors.items():
                all_data[mc_identity][sensor_identity] = sensor_info["data"]

        return json.dumps(all_data, indent=4)



from typing import TypeVar, Type

from ..Utils.Identity import Identity
from ..Config import CERROR_MESSAGES, CCAPABILITIES, CCLIENT
from .Sensor import Sensor
from .Actuator import Actuator
from ..Models.MicrocontrollerCapabilities import MicroControllerCapabilities
from ..Command.CustomCommands import CustomCommands
from ..MayakoData import MayakoData
from ..Utils.Logger import LoggerInstance, LoggerType
from ..Models.BaseDataModel import BaseDataModel

T = TypeVar("T", bound=BaseDataModel) #for sensor model classes
S = TypeVar("S", bound=Actuator) #for actuator classes

class MicroController:

    _identity: str
    _logger: LoggerType

    def __init__(self, mc_identity: str, model: MayakoData, custom_commands: CustomCommands) -> None:
        Identity.register(identity=mc_identity)
        self._logger = LoggerInstance.get()
        self._identity = mc_identity #we store the identity for later reference although it is in capabilities
        self._model = model
        self._commands = custom_commands
        self._check_capabilities(self._identity)

    def _check_capabilities(self, mc_identity: str) -> None:
        """
        checks if the identity provided by the users matches a microcapa in mayakoData

        we do not store the capability here because we want the mayakoData be the single place which is responsible to manage this. also there might be a chance of corruption when the capabilities diverge.

        Args:
            identity (str): the identity of the microcontroller capability
        
        Raises:
            Exception: if no capa was found
        """
        capas = self._model.get_microcontroller_capa_by_identity(identity=mc_identity)
        if not capas:
            raise Exception(f"the microcontroller with the provided identity {self._identity} does not exist in {CCLIENT.CONFIG_FILE_NAME}. add a device with the identity in the mayakoGUI before defining the capabilities.")
        
    def get_capabilities(self) -> MicroControllerCapabilities:
        """retrieves the capabilities of this microcontroller"""
        return self._model.get_microcontroller_capa_by_identity(identity=self._identity)


    def define_capabilities(self, include_timestamp: bool = False, include_sequence: bool = False, delay: int = 0, duration: int = 0, max_samples: int = 0) -> None:
        """
        this method defines the capabilities for the microcontroller.

        first it loads the old capabilities from the microcontroller. then it constructs new capabilities with the information from the users. if the user information does not match with the old, the information in model is updated. pay attention: the old capabilities contain information such as wifi profile, sensors etc. that we want to keep. therefore if a missmatch is, just update the values. the capabilities are not directly updated on the device when the update on the model is performed but when the user executes client.start_record.

        Args:
            include_timestamp (bool): sensor data includes time stamp if true
            include_sequence (bool): sensor data includes a sequence number for each sensor individually
            delay (int): delays the execution of the record for x seconds after the start command was executed; unit seconds; is ignored if 0;
            duration (int): stops the record after x seconds; can be used together with stop command. whichever executes first, stops the record; unit is seconds; ignored if 0;
            max_samples (int): stops the record of x sensor data points were recorded; can run in parallel with duration and stop command; is ignoed if 0;
        """
        old_capas = self._model.get_microcontroller_capa_by_identity(self._identity)
        if not old_capas: #
            raise Exception(f"the microcontroller with the provided identity {self._identity} does not exist in {CCLIENT.CONFIG_FILE_NAME}. add a device with the identity in the mayako-gui before defining the capabilities.")
            
        new_capas = MicroControllerCapabilities(identity=self._identity, include_sequence=include_sequence, include_timestamp=include_timestamp, delay=delay, duration=duration, max_samples=max_samples)

        if not self._compare_capabilities(old_capa=old_capas, new_capa=new_capas):
            #we must update the capabilities
            self._logger.debug("updating microcontroller capabilities")
            #we may only update the changing things!!! the old capa contains infos we still need
            self._update_old_capabilities(old_capa=old_capas, new_capa=new_capas)
            self._model.update_microcontroller_capa(old_capas)
        else:
            self._logger.debug("no update on microcontroller capabilities")

    def _compare_capabilities(self, old_capa: MicroControllerCapabilities, new_capa: MicroControllerCapabilities) -> bool:
        """compares if the microcontroller capabilities match"""
        return (
        old_capa.include_sequence == new_capa.include_sequence and
        old_capa.include_timestamp == new_capa.include_timestamp and
        old_capa.delay == new_capa.delay and
        old_capa.duration == new_capa.duration and
        old_capa.max_samples == new_capa.max_samples
    )

    def _update_old_capabilities(self, old_capa: MicroControllerCapabilities, new_capa: MicroControllerCapabilities) -> None:
        """assign changes from the new microcontroller capability to the old"""
        old_capa.include_sequence = new_capa.include_sequence
        old_capa.include_timestamp = new_capa.include_timestamp
        old_capa.delay = new_capa.delay
        old_capa.duration = new_capa.duration
        old_capa.max_samples = new_capa.max_samples

    def use_sensor(self, sensor_class_type: Type[T], sensor_identity: str) -> Sensor[T]:
        """
        creates a sensor that wants to retrieve sensor data with the type from sensor_class_type

        Args:
            sensor_class_type (Type[T]): a sensor data model class
            sensor_identity (str): identity of a sensor

        Returns:
            Sensor[T]: returns a sensor that 
        """
        sens_capa = self._model.get_sensor_capability_by_identity(self._identity, sensor_identity)
        if not sens_capa:
            raise Exception(f"the sensor with the provided identity {sensor_identity} does not exist in {CCLIENT.CONFIG_FILE_NAME}.")
        
        return Sensor[T](sensor_identity=sensor_identity, mc_identity=self._identity, model=self._model, sensor_class_type=sensor_class_type)

    def use_actuator(self, class_type: Type[S], actuator_identity: str) -> S:
        """
        creates an actuator that can execute commands associated with it

        Args:
            class_type (Type[S]): the specific actuator class which is a subclass of Actuator
            actuator_identity (str): the identity of the actuator

        Return:
            Type[S]: an instance of an actuator implementation which is a sublcass of Actuator
        """
        if not issubclass(class_type, Actuator):
            raise TypeError(f"{class_type.__name__} is not a subclass of {Actuator.__name__}")
        
        act_capa = self._model.get_actuator_capability_by_identity(self._identity, actuator_identity)
        if not act_capa:
            raise Exception(f"the actuator with the provided identity {actuator_identity} does not exist in {CCLIENT.CONFIG_FILE_NAME}.")
        
        return class_type(identity=actuator_identity, mc_identity=self._identity, model=self._model, commands=self._commands)
    
    def get_identity(self) -> str:
        """returns the identity of this microcontroller"""
        return self._identity


from dataclasses import dataclass, asdict
from typing import Dict, Self, Optional
from ..Config import CCLIENT

@dataclass
class User:
    """
    a model class which represents the inidividual machine

    as this project can be used by multiple users, we use this class to distinguish between users as the mac address is unique. this lets users have different arduino folders. uses mayako/Service/MacAddress.py to retrieve it.

    Attributes:
        user_mac_address (str): mac address with following schema: XX:XX:XX:XX:XX:XX
        arduino_folder (str): the path where the arduino project is located; improtant for mayako/Service/ArduinoUploader.py
    """
    user_mac_address: str
    arduino_folder: Optional[str] = CCLIENT.DEFAULT_ARDUINO_FOLDER

    def to_dict(self) -> Dict:
        """makes a dictionary from the class"""
        return asdict(self)
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """returns an instance of the according class by passing a dictionary"""
        default_data = {
            "arduino_folder": CCLIENT.DEFAULT_ARDUINO_FOLDER
        }
        
        merged_data = {**default_data, **data}
        return cls(**merged_data)